#! /bin/env python3
# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import binascii as ba

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class AESprng:
    def __init__(self, seed=list(4*[0])):
        self.reseed(seed) 

    def reseed(self, seed):
        self.key = bytes(12*[0]) + bytes(reversed(seed))
        self.rnd_byte_rem = 0
        self.rnd_buf = bytes(16*[0])
        self.call=0
        self.cipher = Cipher(algorithms.AES(bytes(self.key)), modes.ECB())
        self.cipher = self.cipher.encryptor()


    def nextbyte(self):
        if self.rnd_byte_rem==0:
            self._gen_rnd()
            #print('STATE:',[hex(e) for e in self.rnd_buf])

        rnd_byte = self.rnd_buf[self.rnd_byte_rem-1]
        #print(hex(rnd_byte))
        self.rnd_byte_rem = self.rnd_byte_rem-1
        return rnd_byte

    def clear_buffer(self):
        if self.rnd_byte_rem == 0:
            self._gen_rnd()
        self._gen_rnd()

    def _gen_rnd(self):
        rnd = self.cipher.update(
                self.call.to_bytes(16,byteorder='big')
                )
        self.rnd_buf = rnd
        self.call = self.call+1
        self.rnd_byte_rem = 16


