#! /bin/env python3
# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
def cfg_setup_tt ():
    ram_size = 7
    #ram0 = np.random.randint(0,256,size=7,dtype=np.uint8)
    #ram1 = np.random.randint(0,256,size=7,dtype=np.uint8)
    
    ram0 = np.zeros(ram_size,dtype=np.uint8)
    # Set the '0's
    #ram0[1] = 0
    #ram0[4] = 0
    
    ram1 = ram0

    ram = ram0.tolist() + ram1.tolist()

    # Change 
    flags = [0,0,0,0,0,0,0]

    return dict(
            ram_size = ram_size, # or len(ram0), or the size of a single (parallel) state ...
            ram = ram, # ... But specify the value of ram with the global config, so the concatenation of all parallel state
            flags = flags, 
            nstates = 2,
            base = ram_size # Size of the ram, used by the controller to tap the proper state 
            )
def cfg_sakura_tt_fvr(ram0, ram1, coin):
    ### The data sent to the target is organised as follows:
    ## [rnd, 0, x0, y0, 0, x1, y1]
    ram_size = 7

    ### Define the "ram" for the two config. Basically, all random (but different)
    ### for the two config, with well located zeros to avoid transition
    ### Here, one ram configuration per parallel state, but all of the same length

    # Set the '0's
    ram0[1] = 0
    ram0[4] = 0
    ram1[1] = 0
    ram1[4] = 0
    # Concat the rams. So here, we have two rams of ram_size=7
    if coin == 0:
        ram = ram0.tolist() + ram1.tolist()
    else :
        ram = ram1.tolist() + ram0.tolist()
    ### Generate the flags. Should be the same length as a single ram (i.e., ram_size). 
    ### The same flags will be used in order to generate the data sent from the controller to the target for a single execution.
    ### The same flags are thus used for every execution, independently of the state (randomly) selected. Put in another way,
    ### it is not possible to set different flags configurations for the different parallel 'rams'. 
    ### TODO: verify if we did not implemented that, because we did it for the CW305. If not, do it.
    
    # Here, we basically need that the randomnes is random, and the rest should be fixed
    # Even if we have a state that should be random, this is required to keep the one that need 
    # to be fixed fixed. In practice: legacy ... and should be implemented with a flags configuration independant for 
    # each state (as CW305). 
    flags = [1,0,0,0,0,0,0]

    ### Finally, generate the configuration for the refresh generation. 
    ### Here, it is used only to refresh the sharings of (constant) inputs. 
    ### We do it by:
    ###     - adding the same random constant to x0 and x1
    ###     - adding the same random constant to y0 and y1
    ###     - additionally to fvf: generate a fresh value for the random state
    ### the refresh should be only for first state. We can do it by specifying the following pair of adresses to which 
    ### the random value drawn should be added.

    # Here, resfresh x and y
    r_ram_refresh_x_s0 = [2,5] 
    r_ram_refresh_y_s0 = [3,6] 
    r_ram_s0 = r_ram_refresh_x_s0 + r_ram_refresh_y_s0

    # Here, refresh mechanism use to generate randomness. 
    # Here, solution is to "rotate add" random bytes on each shares
    r_ram_rnd = [2,3,3,5,5,6,6,2] 
    # Global refresh config
    r_ram = r_ram_s0 + r_ram_rnd 
    

    # Return the config as a dict (to allow automatic expension when instanciating the controller)
    return dict(
            ram_size = ram_size, # or len(ram0), or the size of a single (parallel) state ...
            ram = ram, # ... But specify the value of ram with the global config, so the concatenation of all parallel state
            flags = flags, 
            nstates = 2,
            refresh_ram = r_ram, 
            n_refresh = len(r_ram)//2,
            base = ram_size # Size of the ram, used by the controller to tap the proper state 
            )
