#! /bin/env python3
# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

class LCG:
    def __init__(self, a=1664525, c=1013904223):
        self.a = a
        self.c = c

    def reseed(self, seed):
        self.state = int.from_bytes(seed, byteorder='little')
        self.stored = []
        #self.nextbyte() # patch for bug in HW FSM # -> kinda fixed ;)

    def nextbyte(self):
        if not self.stored:
            self.stored = self._advance()
        res = self.stored[0]
        self.stored = self.stored[1:]
        return res

    def _advance(self):
        a = self.state
        self.state = (self.a*self.state + self.c) % 2**32
        #print(f'old state: 0x{a:x}, new state: 0x{self.state:x}')
        return self.state.to_bytes(4, 'big')

