#! /bin/env python3
# SPDX-FileCopyrightText: SIMPLE-Crypto contributors
# SPDX-License-Identifier: Apache-2.0
#
# Copyright SIMPLE-Crypto contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#! /usr/bin/env python

from API_sakura.controller import Controller
from API_sakura.aes_prng import AESprng
from API_sakura.scascope import SCAScope, ChannelConfig
from scalib.metrics import Ttest
from utils import *

from config import *
from controller import Controller, InputByteConfig, InputConfig, BatchConfig  
from tqdm import tqdm
import time
import matplotlib.pyplot as plt
import zlib
import random
import numpy as np

# Create equal classes to check setup impact on ttest
def setup_fixe ():
    # Fixed config
    cfg0 = InputConfig([
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0) # 0
        ])
    # Fixed config
    cfg1 = InputConfig([
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0) # 0
        ])
    return(cfg0, cfg1) 

def setup_cm_fvf ():
    # Fixed config
    cfg0 = InputConfig([
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0xFF), # X0
        InputByteConfig.constant(0xFF), # Y0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0xFF), # X1
        InputByteConfig.constant(0xFF), # Y1
        ])
    # Random config
    cfg1 = InputConfig([
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # X0
        InputByteConfig.constant(0), # Y0
        InputByteConfig.constant(0), # 0
        InputByteConfig.constant(0), # X1
        InputByteConfig.constant(0), # Y1
        ])
    return(cfg0, cfg1) 

def setup_cm_setup ():
    # Fixed config
    cfg0 = InputConfig([
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        ])
    # Random config
    cfg1 = InputConfig([
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        InputByteConfig.constant(0xA3), # 0
        ])
    return(cfg0, cfg1)

def setup_cm_fvr ():
    # Fixed config
    cfg0 = InputConfig([
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(0xFF, 0), # X0
        InputByteConfig.share(0x00, 1), # Y0
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(0x00, 0), # X1
        InputByteConfig.share(0x00, 1) # Y1
        ])
    # Random config
    cfg1 = InputConfig([
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # RND
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(0xFF, 0), # X0
        InputByteConfig.random(), # Y0
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(0x00, 0), # X1
        InputByteConfig.random() # Y1
        ])
    return(cfg0, cfg1) 
# Create equal classes to check setup impact on ttest
def setup_tt ():
    ram_size = 7
    ram0 = np.random.randint(0,256,size=7,dtype=np.uint8)
    #ram1 = np.random.randint(0,256,size=7,dtype=np.uint8)
    
    #ram0 = np.zeros(ram_size,dtype=np.uint8)
    # Set the '0's
    #ram0[1] = 0
    #ram0[4] = 0
    
    ram1 = ram0

    ram = ram0.tolist() + ram1.tolist()

    # Change 
    flags = [0,0,0,0,0,0,0]

    return dict(
            ram_size = ram_size, # or len(ram0), or the size of a single (parallel) state ...
            ram = ram, # ... But specify the value of ram with the global config, so the concatenation of all parallel state
            flags = flags, 
            nstates = 2,
            base = ram_size # Size of the ram, used by the controller to tap the proper state 
            )

def hpc4 (ina, inb, rnd):
    ina0 = ina[1]
    inb0 = inb[1]

    r2 = (rnd&16)>>5
    r3 = (rnd&8)>>4

    ina1 = ina[0]
    inb1 = inb[0]

    out0 = ina0 & inb0 ^ r2 ^ r3 ^ ina0 & inb1
    out1 = ina1 & inb1 ^ r2 ^ r3 ^ ina1 & inb0
    if VERBOSE:
        print('X0', ina0)
        print('X1', ina1)
        print('Y0', inb0)
        print('Y1', inb1)
        print('RND', r2)
        print('RND', r3)
        print('Z0', out0)
        print('Z1', out1)

    return out0, out1

def predict_output (datas_in):
    n = len(datas_in)
    datas_out = np.zeros((n, 7), dtype=np.uint8)
    cnt = 0
    for d in datas_in:
        data_in = d[1]
        rnd = data_in[0]&31
        sx0 = data_in[2]&1
        sx1 = data_in[5]&1
        x = sx0, sx1

        sy0 = data_in[3]&1
        sy1 = data_in[6]&1
        y = sy0, sy1

        sz0, sz1 = hpc4(x, y, rnd)
        cnt += 1
    datas_out = np.array([datas_in[e][1] for e in range(len(datas_in))],dtype=np.uint8)
    return datas_out
# Recombine all values
def recombine(rams):
    xs = rams[:,2] ^ rams[:,5]
    ys = rams[:,3] ^ rams[:,6]
    rs = rams[:,0]
    return rs, xs, ys

def sakura_tt_fvf():
    ### The data sent to the target is organised as follows:
    ## [rnd, 0, x0, y0, 0, x1, y1]
    ram_size = 7

    ### Define the "ram" for the two config. Basically, all random (but different)
    ### for the two config, with well located zeros to avoid transition
    ### Here, one ram configuration per parallel state, but all of the same length
    ram0 = np.random.randint(0,256,size=7,dtype=np.uint8)
    ram1 = np.random.randint(0,256,size=7,dtype=np.uint8)
    # Set the '0's
    ram0[1] = 0
    ram0[4] = 0
    ram1[1] = 0
    ram1[4] = 0
    # Concat the rams. So here, we have two rams of ram_size=7
    ram = ram0.tolist() + ram1.tolist()

    ### Generate the flags. Should be the same length as a single ram (i.e., ram_size). 
    ### The same flags will be used in order to generate the data sent from the controller to the target for a single execution.
    ### The same flags are thus used for every execution, independently of the state (randomly) selected. Put in another way,
    ### it is not possible to set different flags configurations for the different parallel 'rams'. 
    
    # Here, we basically need that the randomnes is random, and the rest should be fixed
    flags = [1,0,0,0,0,0,0]

    ### Finally, generate the configuration for the refresh generation. 
    ### Here, it is used only to refresh the sharings of (constant) inputs. 
    ### We do it by:
    ###     - adding the same random constant to x0 and x1
    ###     - adding the same random constant to y0 and y1
    ### and for both config. We can do it by specifying the following pair of adresses to which 
    ### the random value drawn should be added.
    # For the first config
    r_ram_refresh_x_s0 = [2,5] 
    r_ram_refresh_y_s0 = [3,6] 
    r_ram_s0 = r_ram_refresh_x_s0 + r_ram_refresh_y_s0

    # For the second config
    r_ram_refresh_x_s1 = [ram_size+e for e in [2,5]] 
    r_ram_refresh_y_s1 = [ram_size+e for e in [3,6]] 
    r_ram_s1 = r_ram_refresh_x_s1 + r_ram_refresh_y_s1
    
    # Global refresh config
    r_ram = tuple(r_ram_s0 + r_ram_s1)

    # Return the config as a dict (to allow automatic expension when instanciating the controller)
    return dict(
            ram_size = ram_size, # or len(ram0), or the size of a single (parallel) state ...
            ram = ram, # ... But specify the value of ram with the global config, so the concatenation of all parallel state
            flags = flags, 
            nstates = 2,
            refresh_ram = r_ram, 
            n_refresh = len(r_ram)//2, # legacy ...
            base = ram_size # Size of the ram, used by the controller to tap the proper state 
            )

def sakura_tt_fvr():

    # Fixed config
    cfg0 = InputConfig([
        InputByteConfig.random(), # RND 
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(np.random.randint(0,256,dtype=np.uint8), 0), # X0, random
        InputByteConfig.share(np.random.randint(0,256,dtype=np.uint8), 1), # Y0, random
        InputByteConfig.constant(0), # 0
        InputByteConfig.share(np.random.randint(0,256,dtype=np.uint8), 0), # X1, random
        InputByteConfig.share(np.random.randint(0,256,dtype=np.uint8), 1), # Y1, random
        ])

    # Random config
    cfg1 = InputConfig([
        InputByteConfig.random(), # RND 
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # X0, random
        InputByteConfig.random(), # Y0, random
        InputByteConfig.constant(0), # 0
        InputByteConfig.random(), # X1, random
        InputByteConfig.random(), # Y1, random
        ])

    return (cfg0, cfg1)
### Execution parameters 
CAPTURE = 1 # Should you capture traces ?
VERBOSE = 0 # Do you want explanative prints ?
DEBUG = 0 # Are you debugging?
PLOT = 1

gadget = 'hpc3'
n_gadget = 1
#cfg = setup_tt()
### Parameters 

compute_CRC = True

# External clock
freq_clock = 1e6
gen_clock = False

# Seed for the two independant PRNG embedded inside the controller.
seed0 = [random.randrange(256) for _ in range(4)]
seed1 = [random.randrange(256) for _ in range(4)]

# PICO parameters
var_channel = 'PS5000A_CHANNEL_A'
var_coupling = 'PS5000A_DC'
var_range = 'PS5000A_10MV'

n_extra = 1
n_d = 5
n_c = 4
n_cycle = 2*n_d+n_c+n_extra # Number of cycles to capture
sample_freq = 500000000
sample_delay = 0
var_resolution = 12
var_model = 'PS5000A'
target_freq = 6*1e6

sample_per_clk = sample_freq/target_freq
acq_samples = int(n_cycle*sample_per_clk)

# Acquisition parameters

if DEBUG :     
    wup = 1 
    batch_size = 1000
    batch_size_eff = wup + batch_size
    total_data = 2000 # Number of executon in a run 
else :
    wup = 1000
    batch_size = 100000
    batch_size_eff = wup + batch_size
    total_data = 10e6 # Number of executon in a run 

n_batch = int(total_data // batch_size) 
coin_save = []

# Instance of the SAKURA controller 

#cfg0, cfg1 = sakura_tt_fvr()




c = Controller(
            spi_clock=2E6,
            nb_runs=batch_size_eff,
            seed0 = seed0,
            seed1 = seed1,
            always_reset=True,
            bytes_crc=11,
            cmp_crc=compute_CRC,
            )

cfg0, cfg1 = setup_cm_fvr()
bcfg0 = BatchConfig.generate([cfg0,cfg1])
bcfg1 = BatchConfig.generate([cfg1,cfg0])

save_ttest = []

for b in tqdm(range(n_batch)) : 
    if CAPTURE :
        if b == 0:
            channelA = ChannelConfig(
                        name=var_channel,
                        coupling=var_coupling,
                        vrange=var_range,
                        voffset = 0.000,
                        enabled=True)
            
            ps = SCAScope(channels=[channelA],
                        batch_size=batch_size_eff,
                        n_samples=acq_samples,
                        sample_freq=sample_freq,
                        trig_delay=sample_delay,
                        serialNumber = b'FP650/064',
                        resolution=var_resolution,
                        model=var_model
                        )
        
        ## Derive Testbench
        Ns = ps.noSamples
        batch  = np.zeros((batch_size_eff, Ns), dtype = np.int16)


    t0 = time.process_time_ns()
    
    # Seed for independant PRNG embedded inside the controller.
    
    seed0 = np.random.randint(0,256,size=4).tolist()
    seed1 = np.random.randint(0,256,size=4).tolist()
        
    coin = np.random.randint(0,2, size=1, dtype=np.uint16)
    coin_vector = np.repeat(coin, batch_size_eff)
    coin_save.append(coin)
   
    if coin == 0 :
        bcfg = bcfg0
    else :
        bcfg = bcfg1

    c.apply_config(bcfg)
    c.update(seed0=seed0, seed1=seed1) 

    if b == 0 :
        testvector_file = "../SRC_sakura/BHVL_SIM/devtv.tv"
        c.generate_testvectors_file(testvector_file)

    if VERBOSE :
        print('Applying configuration \n', coin)
    
    # Simulate the states
    ram_predict = c.predict_ram_inputs()
    rams = np.array([p[1] for p in ram_predict], dtype=np.uint8)
    states = np.array([p[0] for p in ram_predict], dtype=np.uint8) 
    rs, xs, ys = recombine(rams)

    # Compute some stats
    I0 = np.where(states==0)[0]
    I1 = np.where(states==1)[0]
    sets = np.zeros(batch_size_eff, dtype=np.uint16)
    sets[I1] = 1
    sets = sets ^ coin_vector
   
    if DEBUG : 
        # compute umsk values in configuration 0
        am_diff_x_s0 = len(np.unique(xs[I0]))
        am_diff_y_s0 = len(np.unique(ys[I0]))

        # compute umsk values in configuration 1
        am_diff_x_s1 = len(np.unique(xs[I1]))
        am_diff_y_s1 = len(np.unique(ys[I1]))

        ####     
        # Somes stats for the amount of different shares
        am_x0_s0 = len(np.unique(rams[I0,2]))
        am_x1_s0 = len(np.unique(rams[I0,5]))
        am_y0_s0 = len(np.unique(rams[I0,3]))
        am_y1_s0 = len(np.unique(rams[I0,6]))

        am_x0_s1 = len(np.unique(rams[I1,2]))
        am_x1_s1 = len(np.unique(rams[I1,5]))
        am_y0_s1 = len(np.unique(rams[I1,3]))
        am_y1_s1 = len(np.unique(rams[I1,6]))

        ## Display some stats
        if VERBOSE :
            print("Amount config 0: {}".format(len(I0)))
            print("Amount config 1: {}".format(len(I1)))
            print("---- Config 0 ----")
            print("Amount of different x for config: {}".format(am_diff_x_s0))
            print("Amount of different x0 for config: {}".format(am_x0_s0))
            print("Amount of different x1 for config: {}".format(am_x1_s0))
            print()
            print("Amount of different y for config: {}".format(am_diff_y_s0))
            print("Amount of different y0 for config: {}".format(am_y0_s0))
            print("Amount of different y1 for config: {}".format(am_y1_s0))
            print("---- Config 1 ----")
            print("Amount of different x for config: {}".format(am_diff_x_s1))
            print("Amount of different x0 for config: {}".format(am_x0_s1))
            print("Amount of different x1 for config: {}".format(am_x1_s1))
            print()
            print("Amount of different y for config: {}".format(am_diff_y_s1))
            print("Amount of different y0 for config: {}".format(am_y0_s1))
            print("Amount of different y1 for config: {}".format(am_y1_s1))

    #sets = np.random.randint(0, 2, batch_size_eff, dtype=np.uint16)
    # Start picoscope watch
    if CAPTURE :
        if VERBOSE :
            print('PICO on watch \n')
        ps.start_acquire()
    # Start execution and wait for its completion
    c.start_run()
    c.wait_finished()

    if CAPTURE :
        if VERBOSE :
            print('PICO gathering traces \n')
        ps.result_acquire(batch[:,:], channelA)
        batch = remove_mean(batch)
        # PLOTS
        if b == 0:
            ttest = Ttest(Ns, d=1)
        ttest.fit_u(batch[wup:], sets[wup:])
        t = ttest.get_ttest()
        if PLOT : 
            f = plt.figure()
            ax0 = f.add_subplot(3,1,1)
            ax1 = f.add_subplot(3,1,2)
            ax2 = f.add_subplot(3,1,3)
            ax = [ax0, ax1, ax2]
            ax[0].plot(batch[wup:wup+9].T)
            ax[1].plot(np.abs(t.T))
            ax[1].axhline(y=4.5, color='r')

            coin_name = ['0', '1']
            coin_number = [coin_save.count(0), coin_save.count(1)]
            ax[2].bar(coin_name, coin_number)
            ts_total = n_cycle*sample_per_clk
            positions = np.arange(0,ts_total, sample_per_clk)
            
            for p in positions:  
                if p/sample_per_clk < 4 :
                    color = 'r'
                elif p/sample_per_clk > 9 :
                    color = 'r'
                else :
                    color = 'b'
                ax[0].axvline(x=p, ymin = -5000, ymax = 5000,color=color)

            plt.savefig('work/ttest.pdf')
            plt.close()

    if b % 10 == 0 :
        print(b)
        print(b%10)
        print('SAVING')
        save_ttest.append(t)

    # Verify the result
    if not(compute_CRC):
        # First case, verify that CRC32 computed on the fly by the controller
        # corresponds to what is expected. That is, compute the expected CRC32
        # based on the results expectedly returned by the target for each execution
        # based o the predicted value sent. 
        if VERBOSE : 
            print('reading ram...')
        ram_read = c.read_ram()
        print('ram read           ', ram_read)
        print('ram predicted (',ram_predict[-1][0],')', ram_predict[-1][1])

        print('Sent')
        print([hex(r) for r in rams[-1]])
        print('\n')
        print([hex(rr) for rr in ram_read])
        print('\n')

        t1 = time.process_time_ns()
        delta_ts = (t1-t0)/(1e9)
        print('Elapsed time (seconds)', delta_ts)
        tp = batch_size / delta_ts
        print('Throughput (trace / seconds)', tp)
        #assert ram_read == ram_predict[-1][1]

    else:
        # Second case, verify that the result returned by the last execution is the same as expected.
        crc = c.read_crc()
        #print("CRC:",hex(crc))
        
        rams = np.array([ram_predict[e][1] for e in range(len(ram_predict))],dtype=np.uint8)
        crc_ref = zlib.crc32(rams)
        #print("REF CRC:",hex(crc_ref))
        
        t1 = time.process_time_ns()
        delta_ts = (t1-t0)/(1e9)
        print('Elapsed time (seconds)', delta_ts)
        tp = batch_size / delta_ts
        print('Throughput (trace / seconds)', tp)
        assert crc==crc_ref 

    np.savez('work/ttest_'+gadget+str(n_gadget)+'.npz', ttests = save_ttest)
    np.savez('work/traces_'+gadget+str(n_gadget)+'.npz', traces = batch)
