// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module clock_generator
(
    input CLK_OSC,
    input CLK_EXT,
    output CLK_FAST_OSC,
    output CLK_MID_OSC,
    output CLK_SLOW_OSC,
    output CLK_SLOW_EXT
);

// Input buffer
wire bufio_clk_osc;
wire bufio_clk_ext;

BUFIO2 #(
    .DIVIDE(1),
    .DIVIDE_BYPASS("false"),
    .I_INVERT("false"),
    .USE_DOUBLER("false")
) BG_OSC (
    .DIVCLK(bufio_clk_osc),
    .I(CLK_OSC)
);

BUFIO2 #(
    .DIVIDE(1),
    .DIVIDE_BYPASS("false"),
    .I_INVERT("false"),
    .USE_DOUBLER("false")
) BG_EXT (
    .DIVCLK(bufio_clk_ext),
    .I(CLK_EXT)
);


/// PLL to generate the clocks FAST and MID 
wire pll_CLK_FAST_OSC;
wire pll_CLK_MID_OSC;
wire pll_CLK_SLOW_OSC;
wire pll_locked;

wire pll_clk_feedback;

PLL_BASE #(
    .BANDWIDTH("OPTIMIZED"), // "HIGH", "LOW" or "OPTIMIZED"
    .CLKFBOUT_MULT(10), // Multiply value for all CLKOUT clock outputs (1-64)
    .CLKFBOUT_PHASE(0.0), // Phase offset in degrees of the clock feedback output (0.0-360.0).
    // CLKOUT0_DIVIDE - CLKOUT5_DIVIDE: Divide amount for CLKOUT# clock output (1-128)
    .CLKOUT0_DIVIDE(10),
    .CLKOUT1_DIVIDE(40),
    .CLKOUT2_DIVIDE(80),
    // CLKOUT0_DUTY_CYCLE - CLKOUT5_DUTY_CYCLE: Duty cycle for CLKOUT# clock output (0.01-0.99).
    .CLKOUT0_DUTY_CYCLE(0.5),
    .CLKOUT1_DUTY_CYCLE(0.5),
    .CLKOUT2_DUTY_CYCLE(0.5),
    // CLKOUT0_PHASE - CLKOUT5_PHASE: Output phase relationship for CLKOUT# clock output (-360.0-360.0).
    .CLKOUT0_PHASE(0.0),
    .CLKOUT1_PHASE(0.0),
    .CLKOUT2_PHASE(0.0),
    .CLK_FEEDBACK("CLKFBOUT"), // Clock source to drive CLKFBIN ("CLKFBOUT" or "CLKOUT0")
    .COMPENSATION("SYSTEM_SYNCHRONOUS"), // "SYSTEM_SYNCHRONOUS", "SOURCE_SYNCHRONOUS", "EXTERNAL"
    .DIVCLK_DIVIDE(1), // Division value for all output clocks (1-52)
    .REF_JITTER(0.1), // Reference Clock Jitter in UI (0.000-0.999).
    .RESET_ON_LOSS_OF_LOCK("FALSE") // Must be set to FALSE
)
PLL_BASE_inst (
    .CLKFBOUT(pll_clk_feedback), // 1-bit output: PLL_BASE feedback output
    // CLKOUT0 - CLKOUT5: 1-bit (each) output: Clock outputs
    .CLKOUT0(pll_CLK_FAST_OSC),
    .CLKOUT1(pll_CLK_MID_OSC),
    .CLKOUT2(pll_CLK_SLOW_OSC),
    .LOCKED(ctrl_locked), // 1-bit output: PLL_BASE lock status output
    .CLKFBIN(pll_clk_feedback), // 1-bit input: Feedback clock input
    .CLKIN(bufio_clk_osc), // 1-bit input: Clock input
    .RST(1'b0) // 1-bit input: Reset input
);

assign CLK_FAST_OSC = pll_CLK_FAST_OSC;
assign CLK_MID_OSC = pll_CLK_MID_OSC;
assign CLK_SLOW_OSC = pll_CLK_SLOW_OSC;
assign CLK_SLOW_EXT = bufio_clk_ext;

endmodule

