// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module clock_manager_unit
(
    input asyn_nrst,
    input CLK_FAST,
    input CLK_MID,
    input CLK_SLOW_OSC,
    input CLK_EXT,
    input S,
    input CLK_SW,
    output CLK_CTRL,
    output CLK_TARGET
);

// Transition guard controler clock (FAST -> MID)
reg guard_ctrl_F_M_0, guard_ctrl_F_M_1, guard_ctrl_F_M_2, guard_ctrl_F_M_3, guard_ctrl_F_M_4;
always@(negedge CLK_FAST, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_F_M_0 <= 0;
else
    guard_ctrl_F_M_0 <= S;

always@(negedge CLK_FAST, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_F_M_1 <= 0;
else
    guard_ctrl_F_M_1 <= guard_ctrl_F_M_0;

always@(negedge CLK_MID, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_F_M_2 <= 0;
else
    guard_ctrl_F_M_2 <= guard_ctrl_F_M_1;

always@(negedge CLK_TARGET, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_F_M_3 <= 0;
else
    guard_ctrl_F_M_3 <= guard_ctrl_F_M_2;

always@(negedge CLK_TARGET, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_F_M_4 <= 0;
else
    guard_ctrl_F_M_4 <= guard_ctrl_F_M_3;

// Transition guard controler clock (MID -> FAST)
reg guard_ctrl_M_F_0, guard_ctrl_M_F_1, guard_ctrl_M_F_2;
always@(negedge CLK_MID, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_M_F_0 <= 0;
else
    guard_ctrl_M_F_0 <= S;

always@(negedge CLK_MID, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_M_F_1 <= 0;
else
    guard_ctrl_M_F_1 <= guard_ctrl_M_F_0;

always@(negedge CLK_FAST, negedge asyn_nrst)
if(!asyn_nrst)
    guard_ctrl_M_F_2 <= 0;
else
    guard_ctrl_M_F_2 <= guard_ctrl_M_F_1;

// Sub clock enable signals
wire CE_ctrl_FM = (
    (guard_ctrl_F_M_0 == guard_ctrl_F_M_1) && 
    (guard_ctrl_F_M_1 == guard_ctrl_F_M_2) && 
    (guard_ctrl_F_M_2 == guard_ctrl_F_M_3) && 
    (guard_ctrl_F_M_3 == guard_ctrl_F_M_4) && 
    (S==1)
);
wire CE_ctrl_MF = (
    (guard_ctrl_M_F_0 == guard_ctrl_M_F_1) && 
    (guard_ctrl_M_F_1 == guard_ctrl_M_F_2) && 
    (S==0)
);

wire CE_ctrl = S ? CE_ctrl_FM : CE_ctrl_MF;

// Mux ctrl
wire mux_ctrl_out;
BUFGMUX 
#(
    .CLK_SEL_TYPE("SYNC")
)
buf_clk_ctrl
(
    .I0(CLK_FAST),
    .I1(CLK_MID),
    .S(S),
    .O(mux_ctrl_out)
);

BUFGCE cg_ctrl(
    .I(mux_ctrl_out),
    .CE(CE_ctrl),
    .O(CLK_CTRL)
);

// Muxes target 
wire mux_target_out_inner;
BUFGMUX 
#(
    .CLK_SEL_TYPE("SYNC")
)
buf_clk_target_0
(
    .I0(CLK_SLOW_OSC),
    .I1(CLK_MID),
    .S(S || CLK_SW),
    .O(mux_target_out_inner)
);

BUFGMUX 
#(
    .CLK_SEL_TYPE("SYNC")
)
buf_clk_target_1
(
    .I0(CLK_EXT),
    .I1(mux_target_out_inner),
    .S(S || !CLK_SW),
    .O(CLK_TARGET)
);



endmodule

