// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module clock_switcher_ctrl
(
    input asyn_nrst,
    input ctrl_clk,
    input dyn_clk_start_target_tx,
    input dyn_clk_start_ctrl_rx,
    input dyn_clk_start_target_rx,
    input dyn_clk_start_config,
    output S_bufg
);

// Holder of the S flag
reg next_S;
reg S_holder;

always@(negedge asyn_nrst, posedge ctrl_clk)
if(!asyn_nrst)
    S_holder <= 0;
else 
    S_holder <= next_S;

always@(*) begin
    next_S = S_holder;
    if(dyn_clk_start_config)
        next_S = 0;
    else if(dyn_clk_start_ctrl_rx)
        next_S = 0;
    else if(dyn_clk_start_target_tx || dyn_clk_start_target_rx)
        next_S = 1;
end
assign S_bufg = S_holder;

endmodule 
