// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module clock_unit
(
    input CLK_OSC,
    input CLK_EXT,
    input CLK_SW,
    input dyn_clk_start_target_tx,
    input dyn_clk_start_ctrl_rx,
    input dyn_clk_start_target_rx,
    input dyn_clk_start_config,
    output CLK_CTRL,
    output CLK_TARGET,
    output CLK_DEBUG
);

// Sub clock generation
wire clk_fast;
wire clk_mid;
wire clk_slow_osc;
wire clk_slow_ext;

BUFG buf_Deb(
    .I(clk_slow_osc),
    .O(CLK_DEBUG)
);

clock_generator clk_gen_unit(
    .CLK_OSC(CLK_OSC),
    .CLK_EXT(CLK_EXT),
    .CLK_FAST_OSC(clk_fast),
    .CLK_MID_OSC(clk_mid),
    .CLK_SLOW_OSC(clk_slow_osc),
    .CLK_SLOW_EXT(clk_slow_ext)
);

wire startup_rst = 1'b0;

wire S_switch;
clock_switcher_ctrl clk_witch(
    .asyn_nrst(!startup_rst),
    .ctrl_clk(CLK_CTRL),
    .dyn_clk_start_target_tx(dyn_clk_start_target_tx),
    .dyn_clk_start_ctrl_rx(dyn_clk_start_ctrl_rx),
    .dyn_clk_start_target_rx(dyn_clk_start_target_rx),
    .dyn_clk_start_config(dyn_clk_start_config),
    .S_bufg(S_switch)
);

clock_manager_unit cmu(
    .asyn_nrst(!startup_rst),
    .CLK_FAST(clk_fast),
    .CLK_MID(clk_mid),
    .CLK_SLOW_OSC(clk_slow_osc),
    .CLK_EXT(clk_slow_ext),
    .S(S_switch),
    .CLK_SW(CLK_SW),
    .CLK_CTRL(CLK_CTRL),
    .CLK_TARGET(CLK_TARGET)
);

endmodule
