// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
/// Control Status Register
/// Stores WORDS words of WORD_SIZE bits update the word N with input when address is N+OFFSET.
module csr #(
	parameter WORD_SIZE=8,
	parameter WORDS=4,
	parameter AW=8,
	parameter OFFSET=0,
	parameter BYTESWAP=0
)(
    input wire clk,
	 input wire reset_n,
	 input wire write,
	 input wire [AW-1:0] address,
	 input wire [WORD_SIZE-1:0] ctrl_data,
	 output reg [WORDS*WORD_SIZE-1:0] state
    );

	 genvar i;
	 for (i=0; i<WORDS; i=i+1) begin
		always@(posedge clk) begin
			if (~reset_n)
				state[((1-BYTESWAP)*i+BYTESWAP*(WORDS-1-i))*WORD_SIZE +: WORD_SIZE] <= 0;
			else if (write && (address == OFFSET+i))
				state[((1-BYTESWAP)*i+BYTESWAP*(WORDS-1-i))*WORD_SIZE +: WORD_SIZE] <= ctrl_data;
		end
	 end

endmodule
