// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module lcg(
    input wire clk,
    input wire reset_n,
	 input wire enable,
	 input wire load_seed,
    input wire [31:0] seed,
    output wire [31:0] dout,
    output wire done
    );
	 
	 parameter A = 1664525;
	 parameter C = 1013904223;
	
	reg enable_delay;
	
	always@(posedge clk) begin
		if (~reset_n)
			enable_delay <= 1'b0;
		else
			enable_delay <= enable;
	end
	
	assign done = enable_delay;
	
	reg [31:0] state;
	
	wire run = (~enable_delay) & enable;
	
	// lfsr
	always@(posedge clk) begin
		if (~reset_n)
			state <= 0;
		else if (load_seed)
			state <= seed;
		else if (run)
			state <= A*state + C;
	end
	assign dout = state;
	

endmodule
