// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module prng_wrapper (
    input wire clk,
    input wire reset_n,
    input wire [31:0] seed,
    input wire prng_request,
    input wire prng_load_seed,
	 input wire clear_prng_buffer,
    output wire [7:0] dout,
    output reg prng_busy
);

// PRNG_BYTE (integer >0 && <256): number of bytes generated per call by the PRNG
localparam PRNG_BYTE = 16;

reg prng_enable;
wire [PRNG_BYTE*8-1:0] prng_out;
wire prng_done;

reg [63:0] call;
reg [PRNG_BYTE*8-1:0] rnd_buf;
reg [7:0] rnd_byte_rem;
reg [31:0]	prng_seed;

assign dout = rnd_buf[0 +: 8];

reg reset_n_aes;

always @(posedge clk) begin
	if (~reset_n || (prng_busy && prng_done)) reset_n_aes <= 0;
	else reset_n_aes <= 1;
	
end

always @(posedge clk) begin
    if (~reset_n)  begin
        prng_busy <= 1;
        prng_enable <= 0;
	end else if (prng_load_seed) begin
        prng_seed <= seed;
        call <= 0;
        prng_busy <= 1;
        prng_enable <= 1;
    end else if(clear_prng_buffer) begin
			prng_busy <= 1;
			prng_enable <= 1;
	 end else if (prng_busy) begin
        if (prng_done) begin
            rnd_buf <= prng_out;
            prng_busy <= 0;
            rnd_byte_rem <= PRNG_BYTE-1;
            prng_enable <= 0;
            call <= call +1;
        end
    end
    else if (prng_request) begin
        if (rnd_byte_rem == 0) begin
            prng_busy <= 1;
            prng_enable <= 1;
        end else begin
            rnd_byte_rem <= rnd_byte_rem -1;
            rnd_buf <= {8'b0, rnd_buf[8 +: PRNG_BYTE*8-8]};
        end
    end
end

AES128_32bit_Tycho aes(
    .clk(clk),
    .reset_n(reset_n_aes),
    .enable(prng_enable),
    .plaintext({64'b0, call}),
    .ciphertext(prng_out),
    .key({96'b0, prng_seed}),
    .done(prng_done)
);

endmodule
