----------------------------------------------------------------------------------
-- Company: UCL Cyrpto Group
-- Engineer: Davide Bellizia
-- 
-- Create Date:    21:00:26 01/18/2018 
-- Design Name: 
-- Module Name:    AES128_32bit_Rocinante - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity AES128_32bit_Tycho is
    Port ( clk : in  STD_LOGIC;
           reset_n : in  STD_LOGIC;
           enable : in  STD_LOGIC;
           plaintext : in  STD_LOGIC_VECTOR (127 downto 0);
           ciphertext : out  STD_LOGIC_VECTOR (127 downto 0);
           key : in  STD_LOGIC_VECTOR (127 downto 0);
           done : out  STD_LOGIC);
end AES128_32bit_Tycho;

architecture Behavioral of AES128_32bit_Tycho is

--COMPONENTS
-- MixColumns
component MixColumns_4byte is
	port (
			data_in : in std_logic_vector (31 downto 0);
			data_out : out std_logic_vector (31 downto 0)
			);
end component MixColumns_4byte;

-- ShiftRows
component ShiftRows_4byte is
	port (
			data_in : in std_logic_vector (31 downto 0);
			data_out : out std_logic_vector (31 downto 0)
			);
end component ShiftRows_4byte;

-- SubsLayer
component SubsLayer_4byte is
	port (
			data_in : in std_logic_vector (31 downto 0);
			data_out : out std_logic_vector (31 downto 0)
			);
end component SubsLayer_4byte;

-- AddRoundKey
component AddRoundKey_4byte is
	port (
			data_in : in std_logic_vector (31 downto 0);
			key_in : in std_logic_vector (31 downto 0);
			data_out : out std_logic_vector (31 downto 0)
			);
end component AddRoundKey_4byte;

-- KeySched
component KeyScheduler_4byte is
    Port ( clk : in  STD_LOGIC;
           reset_n : in  STD_LOGIC;
			  enable : in STD_LOGIC;
			  en_load : in STD_LOGIC;
           inn_s : in  STD_LOGIC_VECTOR (2 downto 0);
           round_counter : in  STD_LOGIC_VECTOR (3 downto 0);
           key_in : in  STD_LOGIC_VECTOR (127 downto 0);
           key_out : out  STD_LOGIC_VECTOR (31 downto 0));
end component KeyScheduler_4byte;
	
-- CONSTANTS

-- TYPES
type state_AES128 is array (15 downto 0) of std_logic_vector (7 downto 0);
type sr_type is array (7 downto 0) of std_logic_vector (7 downto 0);
type sr_32type is array (3 downto 0) of std_logic_vector (31 downto 0);
type state_FSM_type is (idle, S0bis, S0, S1, S2, S3, S4, S5, S6, S7, S8, S9, S10);


-- SIGNALS
signal ciphertext_int : sr_32type;
signal state: state_FSM_type;
signal nextstate : state_FSM_type;
signal key_4byte : std_logic_vector (31 downto 0);
signal data_xor : std_logic_vector (31 downto 0);
signal data_sbox : std_logic_vector (31 downto 0);
signal data_mc : std_logic_vector (31 downto 0);
signal data_sr : std_logic_vector (31 downto 0);
signal data_sbox_d : std_logic_vector (31 downto 0);
signal inn_s, inn_s_2, inn_rr : std_logic_vector (2 downto 0);
signal round_counter,round_counter_2 : std_logic_vector (3 downto 0) := "0000";
signal mux_in_0, mux_in_1, mux_in_2, mux_in_3 : std_logic_vector (7 downto 0);
signal key_sr : std_logic_vector (31 downto 0);
signal sr_line_0, sr_line_1, sr_line_2, sr_line_3 : sr_type;
signal sr_out_sel_0, sr_out_sel_1, sr_out_sel_2, sr_out_sel_3 : std_logic_vector(2 downto 0);
signal mux_in_sel : std_logic_vector (1 downto 0);
signal enable_out : std_logic;
signal en_load : std_logic;
signal en_inn , en_inn_2  : std_logic;
signal enable_sr : std_logic;
signal round_counter_sig : std_logic;

-- ATTRIBUTES
	
begin

KS: KeyScheduler_4byte
    Port map( 
			  clk => clk,
           reset_n => reset_n,
			  enable => enable,
			  en_load => en_load,
           inn_s => inn_s_2,
           round_counter => round_counter_2,
           key_in => key,
           key_out => key_sr
			  );
en_load <= '1' when (state = S0bis or state = s0 or state = s1 or state = s2 ) else '0';

--shift register
sr_0 : process(clk,reset_n,enable, enable_sr )
begin
	if reset_n = '0' then
		sr_line_0 <= (others => (others => '0'));
	elsif enable = '1' and enable_sr = '1' then
		if rising_edge(clk) then
			sr_line_0(0) <= mux_in_0;
			sr_line_0(7 downto 1) <= sr_line_0(6 downto 0);
		end if;
	end if;
end process;

sr_1 : process(clk,reset_n,enable, enable_sr)
begin
	if reset_n = '0' then
		sr_line_1 <= (others => (others => '0'));
	elsif enable = '1' and enable_sr = '1' then
		if rising_edge(clk) then
			sr_line_1(0) <= mux_in_1;
			sr_line_1(7 downto 1) <= sr_line_1(6 downto 0);
		end if;
	end if;
end process;

sr_2 : process(clk,reset_n,enable, enable_sr)
begin
	if reset_n = '0' then
		sr_line_2 <= (others => (others => '0'));
	elsif enable = '1' and enable_sr = '1' then
		if rising_edge(clk) then
			sr_line_2(0) <= mux_in_2;
			sr_line_2(7 downto 1) <= sr_line_2(6 downto 0);
		end if;
	end if;
end process;

sr_3 : process(clk,reset_n,enable, enable_sr)
begin
	if reset_n = '0' then
		sr_line_3 <= (others => (others => '0'));
	elsif enable = '1' and enable_sr = '1' then
		if rising_edge(clk) then
			sr_line_3(0) <= mux_in_3;
			sr_line_3(7 downto 1) <= sr_line_3(6 downto 0);
		end if;
	end if;
end process;


-- output selection
sr_out_sel_0 <= "100" when round_counter = "0000" else "011";

process(inn_s,round_counter)
begin
	if round_counter = "0000" then
		if inn_s < "011" then
			sr_out_sel_1 <= "011";
		else
			sr_out_sel_1 <= "111";
		end if;
		
		if inn_s < "010" then
			sr_out_sel_2 <= "010";
		else
			sr_out_sel_2 <= "110";
		end if;
		
		if inn_s < "001" then
			sr_out_sel_3 <= "001";
		else
			sr_out_sel_3 <= "101";
		end if;
		--sr_out_sel_1 <= "011" when inn_s < "011" else "111";
		--sr_out_sel_2 <= "010" when inn_s < "010" else "110";
		--sr_out_sel_3 <= "001" when inn_s < "001" else "101";
	else
		if inn_s < "011" then
			sr_out_sel_1 <= "010";
		else
			sr_out_sel_1 <= "110";
		end if;
		
		if inn_s < "010" then
			sr_out_sel_2 <= "001";
		else
			sr_out_sel_2 <= "101";
		end if;
		
		if inn_s < "001" then
			sr_out_sel_3 <= "000";
		else
			sr_out_sel_3 <= "100";
		end if;
		--sr_out_sel_1 <= "010" when inn_s < "011" else "110";
		--sr_out_sel_2 <= "001" when inn_s < "010" else "101";
		--sr_out_sel_3 <= "000" when inn_s < "001" else "100";
	end if;
end process;



--inn_rr <= inn_s when round_counter = "0000" else (inn_s);


--data_sr <= sr_line_0(conv_integer(sr_out_sel_0)) & sr_line_1(conv_integer(sr_out_sel_1)) & sr_line_2(conv_integer(sr_out_sel_2)) & sr_line_3(conv_integer(sr_out_sel_3));
data_sr <= sr_line_0(to_integer(unsigned(sr_out_sel_0))) & sr_line_1(to_integer(unsigned(sr_out_sel_1))) & sr_line_2(to_integer(unsigned(sr_out_sel_2))) & sr_line_3(to_integer(unsigned(sr_out_sel_3)));


ARK: AddRoundKey_4byte
	port map(
			data_in => data_sr,
			key_in => key_sr,
			data_out => data_xor
			);

SL: SubsLayer_4byte
	port map(
			data_in => data_xor,
			data_out => data_sbox
			);
			
process(clk,reset_n,enable, inn_s)
	begin
		if reset_n = '0' then
			data_sbox_d <= (others => '0');
		elsif enable = '1' and inn_s /= "100" then
			if rising_edge(clk) then
				data_sbox_d <= data_sbox;
			end if;
		end if;
	end process;
			
MC: MixColumns_4byte
	port map(
			data_in => data_sbox_d,
			data_out => data_mc
			);
			
process(mux_in_sel, plaintext,data_sbox, data_mc, inn_s)
begin
	case mux_in_sel is
		when "00" =>
			mux_in_0 <= (others => '0');
			mux_in_1 <= (others => '0');
			mux_in_2 <= (others => '0');
			mux_in_3 <= (others => '0');
		when "01" =>
			mux_in_0 <= plaintext(127-(conv_integer(inn_s(1 downto 0))*32) downto 120-(conv_integer(inn_s(1 downto 0))*32));
			mux_in_1 <= plaintext(119-(conv_integer(inn_s(1 downto 0))*32) downto 112-(conv_integer(inn_s(1 downto 0))*32));
			mux_in_2 <= plaintext(111-(conv_integer(inn_s(1 downto 0))*32) downto 104-(conv_integer(inn_s(1 downto 0))*32));
			mux_in_3 <= plaintext(103-(conv_integer(inn_s(1 downto 0))*32) downto 96-(conv_integer(inn_s(1 downto 0))*32));
		when "11" =>
			mux_in_0 <= data_mc(31 downto 24);
			mux_in_1 <= data_mc(23 downto 16);
			mux_in_2 <= data_mc(15 downto 8);
			mux_in_3 <= data_mc(7 downto 0);
		when "10" =>
			mux_in_0 <= data_sbox_d(31 downto 24);
			mux_in_1 <= data_sbox_d(23 downto 16);
			mux_in_2 <= data_sbox_d(15 downto 8);
			mux_in_3 <= data_sbox_d(7 downto 0);
		when others =>
			mux_in_0 <= (others => '0');
			mux_in_1 <= (others => '0');
			mux_in_2 <= (others => '0');
			mux_in_3 <= (others => '0');
	end case;
end process;

process(reset_n,clk)
begin
	if reset_n = '0' then
		state <= idle;
	elsif rising_edge(clk) then
		state <= nextstate;
	end if;
end process;

process(state,round_counter,enable, inn_s)
begin
	case state is
		when idle =>
			if enable = '1' then
				nextstate <= S0bis;
			else
				nextstate <= idle;
			end if;
		when S0bis =>
			nextstate <= S0;
		when S0 =>
			nextstate <= S1;
		when S1 =>
			nextstate <= S2;
		when S2 =>
			nextstate <= S3;
		when S3 =>
			if inn_s = "100" then
				nextstate <= S4;
			else
				nextstate <= S3;
			end if;
		when S4 =>
			if round_counter = "1001" and inn_s = "100"then
				nextstate <= S5;
			else
				nextstate <= S4;
			end if;
		when S5 =>
			nextstate <= S6;
		when S6 =>
			nextstate <= S7;
		when S7 =>
			nextstate <= S8;
		when S8 =>
			nextstate <= S9;
		when S9 =>
			nextstate <= idle;
		when others =>
			nextstate <= idle;
	end case;
end process;

enable_out <= '1' when (state = S5 or state = S6 or state = S7 or state = S8) else '0';
enable_sr <= '0' when (state = idle or state = S0bis) else '1';
done <= '1' when state = S9 else '0';

process(state,round_counter)
begin
	case state is
		when idle =>
			mux_in_sel <= "00";
		when S0bis =>
			mux_in_sel <= "00";
		when S0 =>
			mux_in_sel <= "01";
		when S1 =>
			mux_in_sel <= "01";
		when S2 =>
			mux_in_sel <= "01";
		when S3 =>
			mux_in_sel <= "01";
		when S4 =>
			if round_counter >= "1001" then
				mux_in_sel <= "10";
			else
				mux_in_sel <= "11";
			end if;
		when S5 =>
			mux_in_sel <= "10";
		when S6 =>
			mux_in_sel <= "10";
		when S7 =>
			mux_in_sel <= "00";
		when S8 =>
			mux_in_sel <= "00";
		when S9 =>
			mux_in_sel <= "00";
		when otherS =>
			mux_in_sel <= "00";
	end case;
end process;


process(reset_n,clk,enable,state,inn_s)
	begin
		if reset_n = '0' then
			round_counter <= (others => '0');
		elsif enable = '1' and inn_s = "100" and state = S4 then
			if rising_edge(clk) then
				round_counter <= round_counter + 1;
			end if;
		end if;
	end process;
	
	process(reset_n,clk,enable,state,inn_s_2)
	begin
		if reset_n = '0' then
			round_counter_2 <= (others => '0');
		elsif enable = '1' and inn_s_2 = "100" and state = S4 then
			if rising_edge(clk) then
				round_counter_2 <= round_counter_2 + 1;
			end if;
		end if;
	end process;

en_inn <= '0' when state = idle or state = S0bis else '1';
en_inn_2 <= '0' when state = idle else '1';

process(clk, en_inn, reset_n, enable)
begin
	if reset_n = '0' or enable = '0' then
		inn_s <= "000";
	elsif rising_edge(clk) then
		if en_inn = '1' then
			if inn_s < "100" then
				inn_s <= inn_s + 1;
			else
				inn_s <= "000";
			end if;
		end if;
	end if;
end process;

process(clk, en_inn, reset_n, enable, en_inn_2)
begin
	if reset_n = '0' or enable = '0' then
		inn_s_2 <= "000";
	elsif rising_edge(clk) then
		if en_inn_2 = '1' then
			if inn_s_2 < "100" then
				inn_s_2 <= inn_s_2 + 1;
			else
				inn_s_2 <= "000";
			end if;
		end if;
	end if;
end process;

reg_out : process(reset_n,clk, enable_out)
	begin
		if reset_n = '0' then
			ciphertext_int <= (others => (others => '0'));
		elsif enable_out = '1' then
			if rising_edge(clk) then
				ciphertext_int(0) <= data_xor;--(31 downto 24) & data_xor(1) & data_xor(2) & data_xor(3);
				ciphertext_int(3 downto 1) <= ciphertext_int(2 downto 0);
			end if;
		end if;
	end process;


ciphertext <= ciphertext_int (3)(31 downto 24) & ciphertext_int (0)(23 downto 16) & ciphertext_int (1)(15 downto 8) & ciphertext_int (2)(7 downto 0) &
				  ciphertext_int (2)(31 downto 24) & ciphertext_int (3)(23 downto 16) & ciphertext_int (0)(15 downto 8) & ciphertext_int (1)(7 downto 0) &
				  ciphertext_int (1)(31 downto 24) & ciphertext_int (2)(23 downto 16) & ciphertext_int (3)(15 downto 8) & ciphertext_int (0)(7 downto 0) &
				  ciphertext_int (0)(31 downto 24) & ciphertext_int (1)(23 downto 16) & ciphertext_int (2)(15 downto 8) & ciphertext_int (3)(7 downto 0);
end Behavioral;

