----------------------------------------------------------------------------------
-- Company: UCL Crypto Group
-- Engineer: Davide Bellizia
-- 
-- Create Date:    11:42:29 02/13/2018 
-- Design Name: 
-- Module Name:    KeyScheduler_4byte - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity KeyScheduler_4byte is
    Port ( clk : in  STD_LOGIC;
           reset_n : in  STD_LOGIC;
			  enable : in STD_LOGIC;
			  en_load : in STD_LOGIC;
           inn_s : in  STD_LOGIC_VECTOR (2 downto 0);
           round_counter : in  STD_LOGIC_VECTOR (3 downto 0);
           key_in : in  STD_LOGIC_VECTOR (127 downto 0);
           key_out : out  STD_LOGIC_VECTOR (31 downto 0));
end KeyScheduler_4byte;

architecture Behavioral of KeyScheduler_4byte is

--- COMPONENTS
-- SubsLayer
component SubsLayer_4byte is
	port (
			data_in : in std_logic_vector (31 downto 0);
			data_out : out std_logic_vector (31 downto 0)
			);
end component SubsLayer_4byte;

-- AddRoundKey
component AddRoundKey_4byte is
	port (
			data_in : in std_logic_vector (31 downto 0);
			key_in : in std_logic_vector (31 downto 0);
			data_out : out std_logic_vector (31 downto 0)
			);
end component AddRoundKey_4byte;


--- TYPES
type byte_array is array (0 to 9) of std_logic_vector (7 downto 0);
type state_AES_key is array (0 to 3) of std_logic_vector (31 downto 0);

--- CONSTANTS
constant RCON : byte_array := ( x"01",x"02",x"04",x"08",x"10",x"20",x"40",x"80",x"1B",x"36");

--- SIGNALS
signal state_key,state_key_temp : state_AES_key := (others => (others => '0'));
signal F_word : std_logic_vector (31 downto 0);
signal R_word : std_logic_vector (31 downto 0);
signal rot_State : std_logic_vector (31 downto 0);
signal xor_A : std_logic_vector (31 downto 0);
signal xor_B : std_logic_vector (31 downto 0);
signal xor_output : std_logic_vector (31 downto 0);
signal r_cnt_int : std_logic_vector (3 downto 0);

--- ATTRIBUTES

begin

--
process(inn_s, en_load, state_key)
begin
	if en_load = '0' then
			case inn_s is
				when "001" =>
					key_out <= state_key_temp(0)(31 downto 24) & state_key_temp(1)(23 downto 16) & state_key_temp(2)(15 downto 8) & state_key_temp(3)(7 downto 0);
				when "010" =>
					key_out <= state_key_temp(1)(31 downto 24) & state_key_temp(2)(23 downto 16) & state_key_temp(3)(15 downto 8) & state_key_temp(0)(7 downto 0);
				when "011" =>
					key_out <= state_key_temp(2)(31 downto 24) & state_key_temp(3)(23 downto 16) & state_key_temp(0)(15 downto 8) & state_key_temp(1)(7 downto 0);
				when "000" =>
					--key_out <= state_key_temp(3)(31 downto 24)& state_key_temp(0)(23 downto 16) & state_key_temp(1)(15 downto 8) & state_key_temp(2)(7 downto 0);
					key_out <= (others => '0');
				when "100" =>
					key_out <= state_key_temp(3)(31 downto 24)& state_key_temp(0)(23 downto 16) & state_key_temp(1)(15 downto 8) & state_key_temp(2)(7 downto 0);
				when others =>
					key_out <= (others => '0');
			end case;
	else
		key_out <= (others => '0');
	end if;
end process;

process(clk,reset_n, inn_s)
begin
	if reset_n = '0' then
		state_key_temp <= (others => (others => '0'));
	elsif inn_s = "000" then
		if rising_edge(clk) then
			state_key_temp <= state_key;
		end if;
	end if;
end process;

--
process(clk,enable, reset_n, inn_s, en_load)
begin
	if reset_n = '0' or enable = '0' then
		state_key <= (others => (others => '0'));
	elsif rising_edge(clk) then
		if en_load = '0'  then
			case inn_s is
				when "000" =>
					state_key (0) <= xor_output;
				when "001" =>
					state_key (1) <= xor_output;
				when "010" =>
					state_key (2) <= xor_output;
				when "011" =>
					state_key (3) <= xor_output;
				when others =>
					state_key <= state_key;
			end case;
		else
			case inn_s is
				when "000" =>
					state_key (0) <= key_in(127 downto 96);
				when "001" =>
					state_key (1) <= key_in(95 downto 64);
				when "010" =>
					state_key (2) <= key_in(63 downto 32);
				when "011" =>
					state_key (3) <= key_in(31 downto 0);
				when others =>
					state_key <= state_key;
			end case;		
		end if;
	end if;
end process;

--
process(state_key,inn_s,F_word, en_load)
begin
	if en_load = '0' then
	case inn_s is
		when "000" =>		
			xor_A <= state_key(0);
			xor_B <= F_word;
		when "001" =>
			xor_A <= state_key(0);
			xor_B <= state_key(1);
		when "010" =>
			xor_A <= state_key(1);
			xor_B <= state_key(2);
		when "011" =>
			xor_A <= state_key(2);
			xor_B <= state_key(3);
		when others =>
			xor_A <= (others => '0');
			xor_B <= (others => '0');
	end case;
	else
		xor_A <= (others => '0');
		xor_B <= (others => '0');
		end if;
end process;


AddRoundKey_insn : AddRoundKey_4byte
	port map(
			data_in => xor_A,
			key_in => xor_B,
			data_out => xor_output
			);

rot_State <= state_key(3)(23 downto 16) & state_key(3)(15 downto 0) & state_key(3)(31 downto 24);

SBOX_insn: SubsLayer_4byte	
	port map (
			data_in => rot_State,
			data_out => R_word
			);
			
r_cnt_int <= "1001" when round_counter >= "1010" else round_counter;
			
F_word <= (R_word(31 downto 24) xor RCON(to_integer(unsigned(r_cnt_int)))) & R_word(23 downto 0);

end Behavioral;

