`timescale 1ns/1ps
module tb_sakura_dut_sr();
`include "utils.vh"
`include "ucl_cg_sakura_dut_sr_config.vh"

localparam T = 2.0;
localparam Td = T / 2.0;
localparam Toff = 0.1*T;

reg api_clock;
reg rst;
reg [8*BUFFER_IN_BYTES-1:0] buffer_in;
wire [8*BUFFER_OUT_BYTES-1:0] buffer_out;
wire save_buffer_out;
wire rst_buffer_out;
reg start_execution;
wire done_execution;
wire [3:0] LED_STATUS;
wire [15:0] T_GPIO;

always #Td api_clock=~api_clock; 

// From file
reg reset_before_start;
reg [8*BUFFER_IN_BYTES-1:0] buffer_in_from_file;

// Dut
ucl_cg_sakura_dut_sr #(
    .BUFFER_IN_BYTES(BUFFER_IN_BYTES),
    .BUFFER_OUT_BYTES(BUFFER_OUT_BYTES)
) dut(
    .api_clock(api_clock),
    .rst(rst),
    .buffer_in(buffer_in),
    .buffer_out(buffer_out),
    .save_buffer_out(save_buffer_out),
    .rst_buffer_out(rst_buffer_out),
    .start_execution(start_execution),
    .done_execution(done_execution)
);

// Register use to interact with buffer_out
reg [8*BUFFER_OUT_BYTES-1:0] reg_buffer_out; 
always@(posedge api_clock)
if(rst_buffer_out | rst) begin
    reg_buffer_out <= 0;
end else if (save_buffer_out) begin
    reg_buffer_out <= buffer_out;
end else begin
    reg_buffer_out <= reg_buffer_out;
end

// 
integer id_tv_in;
integer nexec;
integer ret;
integer caseid;
integer i;
initial begin
    $dumpfile ("waveform.vcd");
    $dumpvars (0, tb_sakura_dut_sr);
    // Initiate clock
    api_clock = 0;

    // Offset
    #Toff;

    // Open file to read
    $display("TV file used:");
    $display(`TV_IN);
    id_tv_in = $fopen(`TV_IN, "r");

    // Display some additional stuff
    $display("\nGeneration parameter used");
    $display("BUFFER_IN_BYTES: %d",BUFFER_IN_BYTES);
    $display("BUFFER_OUT_BYTES: %d\n",BUFFER_OUT_BYTES);

    // Fetch the amount of execution
    read_file_header(id_tv_in,nexec);

    // Run the execution
    for(i=0;i<nexec;i=i+1) begin
        // Load the case header
        read_case_header(id_tv_in, caseid);   
        // Load buffer in 
        ret = $fscanf(id_tv_in,"RST = %x\n", reset_before_start); 
        if (ret!=1) error_exit();
        ret = $fscanf(id_tv_in,"BUFFER_IN = %x\n", buffer_in_from_file);
        if (ret!=1) error_exit();
        // If required, reset first 
        if (reset_before_start) begin
            rst = 1;
            #T;
            rst = 0;
            #T;
            $display("Reset done");
        end
        buffer_in = buffer_in_from_file;
        #T;
        start_execution = 1;
        #T;
        start_execution = 0;   
        $display("Start execution");
        while(done_execution == 0) begin
            #T;
        end
        #T;
        $display("Case %d", caseid);
        $display("Buffer_in:\t\t 0x%x", buffer_in);
        $display("Buffer_out:\t\t 0x%x\n", reg_buffer_out);
    end

    $finish();

end



endmodule
