// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-P-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-P v2 (https://ohwr.org/cern_ohl_p_v2.txt).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-P v2 for applicable conditions.

`timescale 1ns / 1ps
(* KEEP = "TRUE", S = "TRUE", KEEP_HIERARCHY = "TRUE", DONT_TOUCH = "TRUE" *)
module MSKand_HPC4 #(parameter d=2) (ina, inb, rnd, clk, out);
localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam one_rand = d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [d-1:0] ina;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [d-1:0] inb;
(* syn_keep = "true", keep = "true", fv_type = "random", fv_count = 0, fv_rnd_lat_0 = 0, fv_rnd_count_0 = and_pini_nrnd *) input [and_pini_nrnd-1:0] rnd;
(* fv_type = "clock" *) input clk;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 1 *) output [d-1:0] out;

genvar i,j;

// unpack vector to matrix --> easier for randomness handling

wire [d-1:0] rnd_mat [d-1:0]; 
wire [d-1:0] rnd_mat1 [d-1:0]; 
wire [d-1:0] rnd_mat2 [d-1:0]; 
wire [d-1:0] rnd_mat3 [d-1:0]; 


for(i=0; i<d; i=i+1) begin: igen
    assign rnd_mat[i][i] = 0; 
    assign rnd_mat1[i][i] = 0;
    assign rnd_mat2[i][i] = 0;
    assign rnd_mat3[i][i] = 0;
    for(j=i+1; j<d; j=j+1) begin: jgen
        assign rnd_mat[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)];
        assign rnd_mat[i][j] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+d*(d-1)/2]; 

        assign rnd_mat1[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+2*d*(d-1)/2];
        assign rnd_mat1[i][j] = rnd_mat1[j][i];

        assign rnd_mat2[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+3*d*(d-1)/2]; 
        assign rnd_mat2[i][j] = rnd_mat2[j][i];

        assign rnd_mat3[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+4*d*(d-1)/2]; 
        assign rnd_mat3[i][j] = rnd_mat3[j][i];
    end
end

for(i=0; i<d; i=i+1) begin: ParProdI
	(* KEEP = "TRUE", S = "TRUE", KEEP_HIERARCHY = "TRUE", DONT_TOUCH = "TRUE" *)
    wire [d-2:0] u_F, v_F, w_F;
	 (* KEEP = "TRUE", S = "TRUE", KEEP_HIERARCHY = "TRUE", DONT_TOUCH = "TRUE" *)
    wire aibi_F;
	 (* KEEP = "TRUE", S = "TRUE", KEEP_HIERARCHY = "TRUE", DONT_TOUCH = "TRUE" *)
    wire ina_F;
    wire u_F_calc;
	 wire red_u_F, red_v_F, red_w_F;
	 wire xor_v_f;
	 wire xor_u_v_f;
	 wire aibi;
	 // aibi = ai & bi
    bin_AND #(.W(1)) AND_aibi(
                .ina(ina[i]),
                .inb(inb[i]),
                .out(aibi)
    );
	
		bin_REG #(.W(1)) REG_inainb (
					 .clk(clk),
					 .in(aibi),
					 .out(aibi_F)
		);
		
		bin_REG #(.W(1)) REG_ina (
					 .clk(clk),
					 .in(ina[i]),
					 .out(ina_F)
		);
		
	 bin_redXOR #(.W(d-1)) redXOR_uF(
        .in(u_F),
        .out(red_u_F)
    );
	 
    bin_AND #(.W(1)) AND_u_F_calc(
                .ina(red_u_F),
                .inb(ina_F),
                .out(u_F_calc)
    );
	 
	 bin_redXOR #(.W(d-1)) redXOR_vF(
        .in(v_F),
        .out(red_v_F)
    );
	 
	 bin_redXOR #(.W(d-1)) redXOR_wF(
        .in(w_F),
        .out(red_w_F)
    );
	 
	 bin_XOR #(.W(1)) XOR_out1(
                .ina(red_v_F),
                .inb(red_w_F),
                .out(xor_v_f)
    );
	
	 bin_XOR #(.W(1)) XOR_out2(
                .ina(xor_v_f),
                .inb(u_F_calc),
                .out(xor_u_v_f)
    );
	 
	 bin_XOR #(.W(1)) XOR_out3(
                .ina(xor_u_v_f),
                .inb(aibi_F),
                .out(out[i])
    );
	 
    for(j=0; j<d; j=j+1) begin: ParProdJ
        if (i != j) begin: NotEq
            localparam j2 = j < i ?  j : j-1;

            //wire u = inb[j] ^ rnd_mat[i][j] ^ rnd_mat1[i][j];
            //wire v = ina[i] & rnd_mat[i][j] ^ rnd_mat2[i][j];
            //wire w = ina[i] & rnd_mat1[i][j] ^ rnd_mat3[i][j];
				
				wire xor_u1, and_v1, and_w1;
				
				 bin_XOR #(.W(1)) XOR_G_u1(
								 .ina(rnd_mat[i][j]),
								 .inb(rnd_mat1[i][j]),
								 .out(xor_u1)
				 );
				 
				 bin_XOR #(.W(1)) XOR_G_u2(
								 .ina(xor_u1),
								 .inb(inb[j]),
								 .out(u)
				 );
				 
				 bin_AND #(.W(1)) AND_G_v1(
								 .ina(rnd_mat[i][j]),
								 .inb(ina[i]),
								 .out(and_v1)
				 );
				 
				 bin_XOR #(.W(1)) XOR_G_v2(
								 .ina(and_v1),
								 .inb(rnd_mat2[i][j]),
								 .out(v)
				 );
				 
				 bin_AND #(.W(1)) AND_G_w1(
								 .ina(rnd_mat1[i][j]),
								 .inb(ina[i]),
								 .out(and_w1)
				 );
				 
				 bin_XOR #(.W(1)) XOR_G_w2(
								 .ina(and_w1),
								 .inb(rnd_mat3[i][j]),
								 .out(w)
				 );
				 
				 
				bin_REG #(.W(d-1)) REG_u (
					 .clk(clk),
					 .in(u),
					 .out(u_F[j2])
				);
				
				bin_REG #(.W(d-1)) REG_v (
					 .clk(clk),
					 .in(v),
					 .out(v_F[j2])
				);
				
				bin_REG #(.W(d-1)) REG_w (
					 .clk(clk),
					 .in(w),
					 .out(w_F[j2])
				);
        end
    end
end

endmodule
