// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module ucl_cg_target_dummy(
    input wire T_CLK,
    input wire T_RST_TARGET_B,
    input wire T_RSTB,
    input wire [7:0] T_ADDR,
    input wire [7:0] T_WDATA,
    output [7:0] T_RDATA,
    input wire T_WRITE,
    input wire T_VALID,
    output T_READY,
    input wire T_START,
    output wire [15:0] T_GPIO,
    output wire [3:0] LED_STATUS
);

// Include parameters
`include "ucl_cg_sakura_dut_sr_config.vh"

// API low level handler for transactions
wire [BUFFER_IN_BYTES*8-1:0] buffer_in;
wire [BUFFER_OUT_BYTES*8-1:0] buffer_out;
wire save_buffer_out;
wire start_execution;
wire done_execution;
wire clk;
wire rst;
wire rst_buffer_out;

// Low level API handler. 
ucl_cg_sakura_api_handler #(
    .BUFFER_IN_BYTES(BUFFER_IN_BYTES),
    .BUFFER_OUT_BYTES(BUFFER_OUT_BYTES),
    .DELAY_PIPE_IN(1)
) api_handler (
    .T_CLK(T_CLK),
    .T_RSTB(T_RSTB),
    .T_WDATA(T_WDATA),
    .T_RDATA(T_RDATA),
    .T_WRITE(T_WRITE),
    .T_VALID(T_VALID),
    .T_START(T_START),
    .T_READY(T_READY),
    .buffer_in(buffer_in),
    .buffer_out(buffer_out),
    .save_buffer_out(save_buffer_out),
    .start_execution(start_execution),
    .done_execution(done_execution),
    .api_clock(clk),
    .rst(rst),
    .rst_buffer_out(rst_buffer_out)
);


ucl_cg_sakura_dut_sr #(
    .BUFFER_IN_BYTES(BUFFER_IN_BYTES),
    .BUFFER_OUT_BYTES(BUFFER_OUT_BYTES)
)
dut
(
    .api_clock(clk),
    .rst(rst),
    .buffer_in(buffer_in),
    .buffer_out(buffer_out),
    .save_buffer_out(save_buffer_out),
    .rst_buffer_out(rst_buffer_out),
    .start_execution(start_execution),
    .done_execution(done_execution),
    .LED_STATUS(LED_STATUS),
    .T_GPIO(T_GPIO)
);

endmodule
