// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module UMSK_stall_abs (
    // CLK
    clk,
    // FLAGS
    flag_stall,            // Sel input
    flag_unstall,
    flag_decrypt,
    // PAYLOADS
    pl_in,
    state_in,
    state_out,
    pl_out
);
input clk;
input flag_stall;
input flag_unstall;
input flag_decrypt;
input [0:64-1] pl_in;
input [0:320-1] state_in;
output [0:320-1] state_out;
output [0:64-1] pl_out;

wire [0:320-1] state_stalled, state_stalled_fwd;

//  stall 
regEn #(.count(320)) reg_stall (clk, flag_stall, state_in, state_stalled);
assign state_stalled_fwd = (flag_unstall) ? state_stalled : state_in;
//  in absorption
assign pl_out = pl_in ^ state_stalled_fwd[0:64-1];
assign state_out[64:320-1] = state_stalled_fwd[64:320-1];
assign state_out[0:64-1] = (flag_decrypt) ? pl_in : pl_out;

endmodule