// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module levelled_tb ();
`define DEBUG

// PARAMETER DEFINITION
parameter d = 2;
parameter and_pini_nrnd = 5*d*(d-1)/2;
parameter sbox_pini_nrnd = 5*and_pini_nrnd;
parameter slayer_pini_nrnd = 64*sbox_pini_nrnd;

reg clk;

// CLOCK DEFINITION
localparam Tclk = 10;
localparam Tclkd = (Tclk/2.0);
always @(*) #(Tclk/2.0) clk <= ~clk;

// Crypto core REG & WIRE
reg flag_core_reset, flag_core_decrypt, flag_core_last_pt, flag_core_ad, flag_core_confidentiality, flag_core_integrity;
reg flag_core_in_spa_valid, flag_core_in_dpa_valid, flag_core_out_ready;
reg [0:64-1] in;
reg [0:80-1] pl_core_seed;
reg [0:128-1] pl_core_nonce, k;

wire flag_core_in_spa_ready, flag_core_in_dpa_ready, flag_core_out_valid, flag_core_done;
wire [0:64*d-1] pl_core_spa, pl_core_comp_spa;
wire [0:128*d-1] pl_core_k;
wire [0:128-1] pl_core_tag;

MSKcst #(.d(d), .count(128)) cst_k (k, pl_core_k);
MSKcst #(.d(d), .count(64)) cst_in (in, pl_core_spa);
//assign pl_core_spa = in;

// Crypto core 
levelled_FSM #(.d(d)) core_fsm(
    // CLK
    clk,            // Clock input
    // FLAG SIGNALS
    flag_core_reset,          // flag_core_reset signal
    flag_core_decrypt,        // Decrypt flag
    flag_core_last_pt,
    flag_core_ad,        // Flag to indicate that the spa input is AD
    flag_core_confidentiality,
    flag_core_integrity,
    flag_core_done,
    // AXI
    flag_core_in_spa_valid,   // Master is providing valid data for SPA
    flag_core_in_spa_ready, // Slave is ready to process data for SPA
    flag_core_in_dpa_valid,   // Master is providing valid data for DPA
    flag_core_in_dpa_ready,   // Slave is ready to process data for DPA
    flag_core_out_valid,      // Slave is providing valid data
    flag_core_out_ready,      // Master is ready to process data
    // PAYLOAD
    pl_core_spa,             // 128 input plaintext bus
    pl_core_k,           // 128*d input key bus
    pl_core_nonce,       // 128 input nonce bus
    pl_core_seed,
    pl_core_comp_spa,
    pl_core_tag
);
initial begin
	$dumpfile ("waveform.vcd");
    $dumpvars (0, levelled_tb);
    clk = 0;
    flag_core_reset = 1;
    #20
    flag_core_reset = 0;
    flag_core_last_pt = 0;
    flag_core_decrypt = 0;
    flag_core_confidentiality = 0;
    flag_core_in_dpa_valid = 1;
    flag_core_in_spa_valid = 1;
    flag_core_out_ready = 1;
    k = 128'h0;
    in = 64'h0;
    pl_core_nonce = 128'h0;
    pl_core_seed = 80'h0;
    #500
    flag_core_last_pt = 1;
    #500
    $finish();
end

always@(flag_core_done) begin
	if(flag_core_done) begin
		//#10
		//$finish();
	end
end

endmodule
