// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module DPA_Ascon_tb ();
    //`include "MSKand_HPC2.vh"
    localparam and_pini_mul_nrnd = d*(d-1)/2;
    localparam and_pini_nrnd = and_pini_mul_nrnd;
parameter d = 2;
parameter W = 64;
parameter L = 16;
parameter B = 80;

// WIRE
wire [320*d-1:0] out;

// REGS
reg clk;
reg start_dut, reset, started, perm_a, last;
reg [16*5*and_pini_nrnd-1:0] rnd;
reg [10:0] counter;
reg [320*d-1:0] in;

wire [319:0] out_umsk;
// Clock generation
localparam Tclk = 10;
localparam Tclkd = (Tclk/2.0);
always @(*) #(Tclk/2.0) clk <= ~clk;

// Module instantiation
//MSKcst #(.d(d), .count(320)) cst_in (in_umsk, in);
DPA_Ascon_FSM_d2 #(d) dut (start_dut, clk, reset, perm_a, last, rnd, in, done, out);

// Simulation
initial begin
`ifdef VCD_PATH
                $dumpfile(`VCD_PATH);
`else
               $dumpfile("waveform.vcd");
`endif
		$dumpvars(0, DPA_Ascon_tb);
    clk = 0;
    reset = 1;
    perm_a = 1;
    last = 0;
    #40
    reset = 0;
    start_dut = 1;
    started = 1;
    counter = 0;
    in = 640'h8000800000A000800000000000000000000000020008000A002000220028002A008000820088008A00A000A200A800AA000000020008000A002000220028002A008000820088008A00A000A200A800AA;
    rnd = {(16*5*and_pini_nrnd){1'b1}};
    #10 
    start_dut = 0;
    //in = 640'hX;
    #700
    perm_a = 0;
    #100
    perm_a = 1;
    #400
    last = 1;
end

// Cycle counter for timing purpose
always @(posedge clk) begin
    if(started) begin
        counter = counter + 1;
    end

end

always@(done) begin
	if(done & last) begin
		#20
		$finish();
	end
end
// Debug
  genvar j;
    generate
        for(j=0; j<320; j=j+1) begin
            assign out_umsk[j] = ^(out[d*(j+1)-1:d*j]);
        end
    endgenerate

endmodule