// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module DPA_rc_gen
 (
    clk,
    sel,
    en,
    outmux
);
    input sel, en, clk;
    output wire [7:0] outmux;

    wire [7:0] init, rounding_cst, rounded_cst;
    reg [7:0] rounding_muxed;
    reg [3:0] a, b, a_new, b_new;

    assign init = 8'hF0;

    assign rounding_cst = (sel) ? init : rounded_cst;

    assign outmux = (en) ? rounding_cst : rounding_muxed;
    always @ (posedge clk) begin
        rounding_muxed <= outmux;
            a <= rounding_muxed[7:4];
            b <= rounding_muxed[3:0];
            a_new <= a - 1;
            b_new <= b + 1;
    end

  assign rounded_cst[8-1:4] = a_new;
  assign rounded_cst[3:0] = b_new;

endmodule