// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module DPA_Ascon_datapath #(parameter d = 2) (
    clk, 
    flag_dp_start, 
    flag_dp_perm, 
    pl_dp_in, 
    pl_dp_rnd, 
    pl_dp_out
    );

parameter L = 16;
input clk, flag_dp_start, flag_dp_perm;
input [320*d-1:0] pl_dp_in;
input [64*25*d*(d-1)/2-1:0] pl_dp_rnd;

output [320*d-1:0] pl_dp_out;

reg [320-1:0] outmux_f;
wire [320-1:0] outmux, u_state_csl, u_state_cs, u_state_c;
wire en;
wire[8-1:0] cst;
wire[8*d-1:0] m_cst;
wire [320*d-1:0] state_rounding, state_rounding_c, state_rounding_cs;

MSKmux #(.d(d), .count(320)) mux_state (flag_dp_start, pl_dp_in, pl_dp_out, state_rounding);

DPA_rc_gen cstgeneration (clk, flag_dp_start, flag_dp_perm, cst);

MSKcst #(.d(d), .count(8)) msk_state (cst, m_cst);

// P_c

assign state_rounding_c[128*d +: 8*d] = state_rounding[128*d +:8*d] ^ m_cst;
assign state_rounding_c[320*d-1:136*d] = state_rounding[320*d-1:136*d];
assign state_rounding_c[128*d-1 : 0] = state_rounding[128*d-1:0];

// P_s

    genvar i;
    generate 
    for(i=0; i<64; i = i + 1) begin: sboxs 
        MSK_Sbox #(.d(d)) sboxi (
            .clk(clk),
            .rnd(pl_dp_rnd[25*d*(d-1)/2*i +: 25*d*(d-1)/2]),
            .in({state_rounding_c[256*d + i*d +: d], state_rounding_c[192*d + i*d +: d], state_rounding_c[128*d + i*d +: d], state_rounding_c[64*d + i*d +: d], state_rounding_c[i*d +: d]}),
            .out({state_rounding_cs[256*d + i*d +: d], state_rounding_cs[192*d + i*d +: d], state_rounding_cs[128*d + i*d +: d], state_rounding_cs[64*d + i*d +: d], state_rounding_cs[i*d +: d]})
        );
    end
    endgenerate
    
// P_l

MSK_PermL #(.d(d)) mskpermL (state_rounding_cs, pl_dp_out);

`ifdef DEBUG
genvar j;
    generate
        for(j=0; j<320; j=j+1) begin
            assign u_state_c[j] = ^(state_rounding_c[d*(j+1)-1:d*j]);
            assign u_state_cs[j] = ^(state_rounding_cs[d*(j+1)-1:d*j]);
            assign u_state_csl[j] = ^(pl_dp_out[d*(j+1)-1:d*j]);
        end
    endgenerate
`endif
endmodule