// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module DPA_tb_ascon ();

// PARAMETER DEFINITION
parameter d = 2;
parameter W = 64;
parameter L = 16;
parameter B = 80;

parameter and_pini_nrnd = 5*d*(d-1)/2;
parameter sbox_pini_nrnd = 5*and_pini_nrnd;
parameter slayer_pini_nrnd = 64*sbox_pini_nrnd;

//`define DEBUG

reg clk;

// CLOCK DEFINITION
localparam Tclk = 10;
localparam Tclkd = (Tclk/2.0);
always @(*) #(Tclk/2.0) clk <= ~clk;

// Debug
wire[0:320-1] out_umsk;

// PRNG Regs and Wires
reg flag_prng_start, flag_prng_ready, flag_prng_reset;
wire flag_prng_busy, flag_prng_valid;
reg [0:80-1] pl_prng_seed;
wire [0:slayer_pini_nrnd-1] pl_prng_rnd;

// DPA Core Regs and Wires
reg flag_dpa_start, flag_dpa_perm, flag_dpa_mask, flag_dpa_reset, flag_dpa_last;
wire flag_dpa_transition, flag_dpa_done;
reg [0:320-1] um_pl_dpa_in;
wire [0:320*d-1] pl_dpa_in, pl_dpa_out;
wire [0:slayer_pini_nrnd-1] pl_dpa_rnd;

prng_top #(.RND(slayer_pini_nrnd), .NINIT(4*288), .MAX_UNROLL(128)) trivium1 (
    flag_prng_reset,
    clk,            // Clock
    flag_prng_start,   // Acive high. flag_dpa_start the reseeding procedure.
    // Active high, reseed in progress,
    flag_prng_busy,
    // The seed, loaded as initial state.
    pl_prng_seed,
    // SVRS pseudo-random output stream.
    // Output will be always valid, except:
    // - when a reseeding is in progress
    // - when the PRNG has not been reseeded yet after a flag_dpa_reset.
    flag_prng_ready,
    flag_prng_valid,
    pl_prng_rnd
    );

MSKcst #(.d(d), .count(320)) state_cst (um_pl_dpa_in, pl_dpa_in);
assign pl_dpa_rnd = pl_prng_rnd;

DPA_Ascon_FSM_d2 dut (
    // CLK
    clk,
    // FLAGS
    flag_dpa_start,
    flag_dpa_perm,
    flag_dpa_mask,
    flag_dpa_reset,
    flag_dpa_last, 
    flag_dpa_transition,
    flag_dpa_done,
    // PAYLOADS
    pl_dpa_in,
    pl_dpa_rnd,
    pl_dpa_out
);

initial begin
	$dumpfile ("waveform.vcd");
    $dumpvars (0, DPA_tb_ascon);
    clk = 0;
    flag_prng_reset = 1;
    flag_dpa_reset = 1;
    flag_dpa_start = 0;
    flag_dpa_perm = 1;
    flag_dpa_mask = 1;
    #10
    flag_prng_start = 1;
    flag_prng_reset = 0;
    flag_dpa_reset = 0;
    pl_prng_seed = 80'h0;
    #10
    flag_prng_start = 0;
    #100
    flag_prng_ready = 1;
    flag_dpa_start = 1;
    um_pl_dpa_in = 320'h80400c06000000003100000000000000000000000000000000000000000000000000000000000000;
    #1000
    $finish();
end


always@(flag_dpa_done) begin
	if(flag_dpa_done & flag_dpa_last) begin
		#20
		$finish();
	end
end

// Debug
`ifdef DEBUG
genvar j;
    generate
        for(j=0; j<320; j=j+1) begin
            assign out_umsk[j] = ^(pl_dpa_out[d*j:d*(j+1)-1]);
        end
    endgenerate
`endif
endmodule