// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module DPA_Ascon_datapath #(parameter d = 2) (
    clk, 
    flag_dp_start, 
    flag_dp_gating,
    flag_dp_perm, 
    pl_dp_in, 
    pl_dp_rnd, 
    pl_dp_out
);
// CLK
input clk;
// FLAGS
input flag_dp_start;
input flag_dp_gating;
input flag_dp_perm;
// PAYLOADS
input [0:64*5*5*d*(d-1)/2-1] pl_dp_rnd;
input [0:320*d-1] pl_dp_in;
output [0:320*d-1] pl_dp_out;

wire [0:8-1] cst;
wire [0:8*d-1] m_cst;
wire [0:320*d-1] pl_rounding, pl_rounded, state_rounding_cs, state_rounding_c;

MSKmux #(.d(d), .count(320)) mux_init (flag_dp_start, pl_dp_in, pl_rounded, pl_rounding);

// P_c

DPA_rc_gen cstgeneration (clk, flag_dp_start, flag_dp_perm, cst);

MSKcst #(.d(d), .count(8)) msk_state (cst, m_cst);

assign state_rounding_c[184*d:192*d-1] = pl_rounding[184*d:192*d-1] ^ m_cst;
assign state_rounding_c[192*d:320*d-1] = pl_rounding[192*d:320*d-1];
assign state_rounding_c[0:184*d-1] = pl_rounding[0:184*d-1];
// P_s

genvar i;
generate 
for(i=0; i<64; i = i + 1) begin: sboxs 
    MSK_Sbox #(.d(d)) sboxi (
        .clk(clk),
        .gating(flag_dp_gating),
        .rnd(pl_dp_rnd[25*d*(d-1)/2*i : 25*d*(d-1)/2*(i+1)-1]),
        .in({state_rounding_c[i*d : (i+1)*d-1], state_rounding_c[64*d + i*d +: d], state_rounding_c[128*d + i*d +: d], state_rounding_c[192*d + i*d +: d], state_rounding_c[256*d + i*d +: d]}),
        .out({state_rounding_cs[i*d +: d], state_rounding_cs[64*d + i*d +: d], state_rounding_cs[128*d + i*d +: d], state_rounding_cs[192*d + i*d +: d], state_rounding_cs[256*d + i*d +: d]})
    );
end
endgenerate
    
// P_l
MSK_PermL #(.d(d)) mskpermL (state_rounding_cs, pl_rounded); 
assign pl_dp_out = pl_rounded;

`ifdef DEBUG
wire [0:320-1] db_state_rounding_c, db_state_rounding_cs, db_state_rounded_csl;
    genvar j;
    generate
        for(j=0; j<320; j=j+1) begin
            assign db_state_rounding_c[j] = ^(state_rounding_c[d*j:d*(j+1)-1]);
        end
    endgenerate
    genvar k;
    generate
        for(k=0; k<320; k=k+1) begin
            assign db_state_rounding_cs[k] = ^(state_rounding_cs[d*k:d*(k+1)-1]);
        end
    endgenerate
    genvar l;
    generate
        for(l=0; l<320; l=l+1) begin
            assign db_state_rounded_csl[l] = ^(pl_rounded[d*l:d*(l+1)-1]);
        end
    endgenerate
`endif
endmodule

