// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module MSK_Sbox #(parameter d = 2) (clk, rnd, gating, in, out);
localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
// INPUT / OUTPUT
input clk;
input [5*and_pini_nrnd-1:0] rnd;
input gating;
input [0:5*d-1] in;
output [0:5*d-1] out;

wire [0:5-1] out_umsk, in_umsk;

// INTERMEDIATE VALUES
wire [0:d-1] s_0_x, s_0_xi, s_0_xia, s_0_xx;
wire [0:d-1] s_1_i, s_1_ia, s_1_x;
wire [0:d-1] s_2_x, s_2_xi, s_2_xx, s_2_xxi, s_2_xia;
wire [0:d-1] s_3_x, s_3_i, s_3_ix;
wire [0:d-1] s_4_x, s_4_xi, s_4_xia;
wire [0:d-1] in1_F, s_1_i_F, s_2_x_F, s_2_xi_F, in3_F, s_0_x_F, in1_FF, s_4_x_F;
// PROCESSING

MSKxor #(d) xor1 (in[0:d-1], in[4*d:5*d-1], s_0_x);
MSKregGated #(d) reg1 (clk, gating, in[d:2*d-1], in1_F);
MSKregGated #(d) reg11 (clk, gating, s_0_x, s_0_x_F);
MSKinv #(d) inv1 (s_0_x, s_0_xi);
MSKand_HPC4_gated #(d) and1 (in[d:2*d-1], s_0_xi, rnd[0 +: and_pini_nrnd], gating, clk, s_0_xia);
MSKxor #(d) xor2 (s_0_x_F, s_1_ia, s_0_xx);
MSKxor #(d) xor3 (s_0_xx, out[4*d:5*d-1], out[0:d-1]);

MSKinv #(d) inv2 (in[d:2*d-1], s_1_i);
MSKand_HPC4_gated #(d) and2 (s_1_i, s_2_x, rnd[and_pini_nrnd +: and_pini_nrnd], gating, clk, s_1_ia);
MSKxor #(d) xor4 (in1_F, s_2_xia, s_1_x);
MSKxor #(d) xor5 (s_1_x, s_0_xx, out[d:2*d-1]);

MSKxor #(d) xor6 (in[d:2*d-1], in[2*d:3*d-1], s_2_x);
MSKinv #(d) inv3 (s_2_x, s_2_xi);
MSKregGated #(d) reg9 (clk, gating, s_2_xi, s_2_xi_F);
MSKxor #(d) xor7 (s_2_x_F, s_3_ix, s_2_xx);
MSKinv #(d) inv4 (s_2_xx, out[2*d:3*d-1]);
MSKregGated #(d) reg3 (clk, gating, s_2_x, s_2_x_F);
MSKand_HPC4_gated #(d) and3 (s_2_xi, in[3*d:4*d-1], rnd[2*and_pini_nrnd +: and_pini_nrnd], gating, clk, s_2_xia);

MSKxor #(d) xor8 (in3_F, s_4_xia, s_3_x);
MSKxor #(d) xor9(s_3_x, s_2_xx, out[3*d:4*d-1]);
MSKinv #(d) inv5(in[3*d:4*d-1], s_3_i);
MSKregGated #(d) reg4 (clk, gating, in[3*d:4*d-1], in3_F);
MSKand_HPC4_gated #(d) and4 (s_3_i, s_4_x, rnd[3*and_pini_nrnd +: and_pini_nrnd], gating, clk, s_3_ix);

MSKxor #(d) xor10 (in[4*d:5*d-1], in[3*d:4*d-1], s_4_x);
MSKregGated #(d) reg42 (clk, gating, s_4_x, s_4_x_F);
MSKinv #(d) inv6 (s_4_x, s_4_xi);
MSKand_HPC4_gated #(d) and5 (s_0_x, s_4_xi, rnd[4*and_pini_nrnd +: and_pini_nrnd], gating, clk, s_4_xia);
MSKxor #(d) xor11 (s_4_x_F, s_0_xia, out[4*d:5*d-1]);

endmodule