// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns/1ps
module PermL (state, update); 
    parameter W = 64;

    input [0 : 5*W-1] state;
    output [0 : 5*W-1] update;

    wire [0 : W-1] x0, x1, x2, x3, x4;
    wire [0 : W-1] x0_p1, x0_p2, x1_p1, x1_p2, x2_p1, x2_p2, x3_p1, x3_p2, x4_p1, x4_p2;
    wire [0 : W-1] x0_modif, x1_modif, x2_modif, x3_modif, x4_modif;

    assign x0 = state[0 : W-1];
    assign x1 = state[W : 2*W-1];
    assign x2 = state[2*W : 3*W-1];
    assign x3 = state[3*W : 4*W-1];
    assign x4 = state[4*W : 5*W-1];

    //L1
    assign x0_p1 = {x0[64-19 : 64-1], x0[0 : 64-20]};
    assign x0_p2 = {x0[64-28 : 64-1], x0[0 : 64-29]};
    assign x0_modif = x0_p1 ^ x0_p2;
    assign update[0 : W-1] = x0 ^ x0_modif;
    //L2
    
    assign x1_p1 = {x1[64-61 : 64-1], x1[0 : 64-62]};
    assign x1_p2 = {x1[64-39 : 64-1], x1[0 : 64-40]};
    assign x1_modif = x1_p1 ^ x1_p2;
    assign update[W : 2*W-1] = x1 ^ x1_modif;
    
    //L3
        
    assign x2_p1 = {x2[64-1], x2[0 : 64-2]};
    assign x2_p2 = {x2[64-6 : 64-1], x2[0 : 64-7]};
    assign x2_modif = x2_p1 ^ x2_p2;
    assign update[2*W : 3*W-1] = x2 ^ x2_modif;
    //L4
    
    assign x3_p1 = {x3[64-10 : 64-1], x3[0 : 64-11]};
    assign x3_p2 = {x3[64-17 : 64-1], x3[0 : 64-18]};
    assign x3_modif = x3_p1 ^ x3_p2;
    assign update[3*W : 4*W-1] = x3 ^ x3_modif;
    //L5
    assign x4_p1 = {x4[64-7 : 64-1], x4[0 : 64-8]};
    assign x4_p2 = {x4[64-41 : 64-1], x4[0 : 64-42]};
    assign x4_modif = x4_p1 ^ x4_p2;
    assign update[4*W : 5*W-1] = x4 ^ x4_modif;
    
endmodule