// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns/1ps
module SPA_Ascon_FSM (
    flag_spa_start,
    clk,
    flag_spa_reset,
    flag_spa_last,
    pl_spa_in,
    flag_spa_done,
    flag_spa_transition,
    pl_spa_out
);

input flag_spa_start, clk, flag_spa_reset, flag_spa_last;
input [0:320-1] pl_spa_in;
output reg flag_spa_transition, flag_spa_done;
output [0:320-1] pl_spa_out;

reg sel, reset_round_cnt;
reg [2:0] round_cnt;
reg state, next_state;

SPA_Ascon_datapath datapath(sel, clk, pl_spa_in, pl_spa_out);

// STATE DEFINITION
`define IDLE 0
`define COMPUTE 1

always@(*) begin
    next_state = state;
    sel = 0;
    reset_round_cnt = 0;
    flag_spa_done = 0;
    flag_spa_transition = 0;
    case (state) 
        `IDLE : begin
            reset_round_cnt = 1;
            if(flag_spa_start) begin
                sel = 1;
                next_state = `COMPUTE;
                reset_round_cnt = 1;
            end
        end
        `COMPUTE : begin
            if (round_cnt == 4) begin
                flag_spa_transition = 1;
            end
            if (round_cnt == 5) begin
                reset_round_cnt = 1;
                flag_spa_done = 1;
                if (flag_spa_last) begin
                    next_state = `IDLE;
                end
                else begin
                    sel = 1;
                end
            end
        end
    endcase
end

// STATE EVOLUTION
always @(posedge clk) begin
    if (flag_spa_reset) begin
        state <=  `IDLE;
    end
    else begin
        state <= next_state;
    end
end 
// ROUND COUNTER EVOLUTION
always @(posedge clk) begin
    if(reset_round_cnt) begin
        round_cnt <= 0;
    end else begin
        round_cnt <= round_cnt + 1;
    end
end

endmodule