// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns/1ps
module SPA_Ascon_datapath (
    sel, 
    clk, 
    in, 
    out
    );

input sel, clk;
input [0:320-1] in;
output [0:320-1] out;

// PARAMETERS
parameter W = 64;

wire [0:320-1] state_c, state_cs, state_csl, state_in;
reg [0: 320-1] state_rounded;
wire [0:8-1] rnd_cst;

assign state_in = (sel) ? in : state_rounded;

SPA_rc_gen cstgen (clk, sel, 1'b0, rnd_cst);

// PERMUTATION C
assign state_c[0:183] = state_in[0:183];
assign state_c[184:191] = state_in[184:192-1] ^ rnd_cst;
assign state_c[192:319] = state_in[192:319];
// PERMUTATION S
genvar i;
generate 
for(i=0; i<64; i = i + 1) begin: sboxs 
    Sbox sboxi (
        .clk(clk),
        .in({state_c[i], state_c[W + i], state_c[2*W + i], state_c[3*W + i], state_c[4*W + i]}),
        .out({state_cs[i], state_cs[W + i], state_cs[2*W + i], state_cs[3*W + i], state_cs[4*W + i]})
    );
end
endgenerate
// PERMUTATION L
PermL permL (state_cs, state_csl);
always@(posedge clk) begin
    state_rounded <= state_csl;
end
assign out = state_rounded;
endmodule