// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns/1ps
module SPA_Ascon_datapath (
    sel, 
    clk, 
    in, 
    out
    );

input sel, clk;
input [0:320-1] in;
output [0:320-1] out;

// PARAMETERS
parameter W = 64;

wire [0:320-1] state_c, state_cs, state_rounded1, state_r_c, state_r_cs, state_r_csl, state_rr_c, state_rr_cs, state_rr_csl, state_in;
reg [0: 320-1] state_rounded3;
wire [0:8-1] rnd_cst1, rnd_cst2, rnd_cst3;

// PERM1

assign state_in = (sel) ? in : state_rounded3;

SPA_rc_gen cstgen1 (clk, sel, 8'h96, rnd_cst1);

// PERMUTATION C
assign state_c[0:183] = state_in[0:183];
assign state_c[184:191] = state_in[184:192-1] ^ rnd_cst1;
assign state_c[192:319] = state_in[192:319];
// PERMUTATION S
genvar i;
generate 
for(i=0; i<64; i = i + 1) begin: sboxs1
    Sbox sboxi (
        .clk(clk),
        .in({state_c[i], state_c[W + i], state_c[2*W + i], state_c[3*W + i], state_c[4*W + i]}),
        .out({state_cs[i], state_cs[W + i], state_cs[2*W + i], state_cs[3*W + i], state_cs[4*W + i]})
    );
end
endgenerate
// PERMUTATION L
PermL permL1 (state_cs, state_rounded1);


// PERM 2
SPA_rc_gen cstgen2 (clk, sel, 8'h87, rnd_cst2);

// PERMUTATION C
assign state_r_c[0:183] = state_rounded1[0:183];
assign state_r_c[184:191] = state_rounded1[184:192-1] ^ rnd_cst2;
assign state_r_c[192:319] = state_rounded1[192:319];
// PERMUTATION S
genvar j;
generate 
for(j=0; j<64; j = j + 1) begin: sboxs2 
    Sbox sboxi (
        .clk(clk),
        .in({state_r_c[j], state_r_c[W + j], state_r_c[2*W + j], state_r_c[3*W + j], state_r_c[4*W + j]}),
        .out({state_r_cs[j], state_r_cs[W + j], state_r_cs[2*W + j], state_r_cs[3*W + j], state_r_cs[4*W + j]})
    );
end
endgenerate
// PERMUTATION L
PermL permL2 (state_r_cs, state_r_csl);

// PERM3
SPA_rc_gen cstgen3 (clk, sel, 8'h78, rnd_cst3);

// PERMUTATION C
assign state_rr_c[0:183] = state_r_csl[0:183];
assign state_rr_c[184:191] = state_r_csl[184:192-1] ^ rnd_cst3;
assign state_rr_c[192:319] = state_r_csl[192:319];
// PERMUTATION S
genvar k;
generate 
for(k=0; k<64; k = k + 1) begin: sboxs3
    Sbox sboxi (
        .clk(clk),
        .in({state_rr_c[k], state_rr_c[W + k], state_rr_c[2*W + k], state_rr_c[3*W + k], state_rr_c[4*W + k]}),
        .out({state_rr_cs[k], state_rr_cs[W + k], state_rr_cs[2*W + k], state_rr_cs[3*W + k], state_rr_cs[4*W + k]})
    );
end
endgenerate
// PERMUTATION L
PermL permL3 (state_rr_cs, state_rr_csl);
always@(posedge clk) begin
    state_rounded3 <= state_rr_csl;
end
assign out = state_rounded3;
endmodule