`timescale 1ns / 1ps
`ifndef DEFAULTSHARES
`define DEFAULTSHARES 2
`endif
module DPA_Ascon_FSM #(parameter d = `DEFAULTSHARES, parameter width = 16) (
    clk,
    start,
    flag_dp_perm_a,
    reset,
    last, 
    state_in,
    key_in,
    rnd,
    state_out,
    flag_transition,
    done
);


    // INPUT
    input clk, start, flag_dp_perm_a, reset, last;
    input [0:width*3*d-1] state_in;
    input [0:width*2*d-1] key_in;
    input [0:width*25*d*(d-1)/2-1] rnd;
    // OUTPUT
    output [0:width*5*d-1] state_out;
    output reg flag_transition, done;

    // REG & WIRE
    reg start_round, reset_round_cnt, sel, sel_cst;
    reg [0:1] state, next_state;
    reg [0:3] round_cnt;
    wire [0:width*5*d-1] state_dpa;

    // STATE DEFINITION
    `define IDLE 0
    `define PERM_A 1
    `define PERM_B 2

    assign state_dpa = {state_in, key_in};

    // INSTANTIATION
    DPA_Ascon_datapath #(.d(d), .width(width)) datapath (clk, sel, sel_cst, state_dpa, rnd, state_out);

always @(*) begin
    next_state = state;
    sel = 0;
    done = 0;
    start_round = 0;
    reset_round_cnt = 0;
    sel_cst = 0;
    flag_transition = 0;
    case (state)
        `IDLE : begin
            reset_round_cnt = 1;
            if (start) begin
                if(flag_dp_perm_a) begin
                    next_state = `PERM_A;
                    sel = 1;
                    sel_cst = 1;
                    reset_round_cnt = 1;
                end
                else begin
                    next_state = `PERM_B;
                    sel = 1;
                    reset_round_cnt = 1;
                end
            end
        end
        `PERM_A : begin
            if (round_cnt == 10) begin
                flag_transition = 1;
            end
            if(round_cnt == 11) begin
                done = 1;
                reset_round_cnt = 1;
                sel = 1;
                if(last) begin
                    next_state = `IDLE;
                end
                else if(~flag_dp_perm_a) begin
                    next_state = `PERM_B;
                end
                else begin
                end
            end
        end
        `PERM_B : begin
            if (round_cnt == 4) begin
                flag_transition = 1;
            end
            if(round_cnt == 5) begin
                done = 1;
                reset_round_cnt = 1;
                sel = 1;
                if(flag_dp_perm_a) begin
                    next_state = `PERM_A;
                end
            end
        end
    endcase
end

always @(posedge clk) begin
    if (reset) begin
        state <=  `IDLE;
    end
    else begin
        state <= next_state;
    end
end 

always @(posedge clk) begin
    if(reset_round_cnt) begin
        round_cnt <= 0;
    end
    else begin
        round_cnt <= round_cnt + 1;
    end
end

endmodule