`timescale 1ns / 1ps
module DPA_tb_ascon ();
// Select desired gadget
`define and_gate(ina, inb, rnd, clk, out) MSKand_HPC4(ina, inb, rnd, clk, out)
// PARAMETER DEFINITION
parameter d = 2;
parameter W = 64;
parameter L = 16;
parameter B = 80;
parameter width = 8;
parameter and_pini_nrnd = 5*d*(d-1)/2;

// CLOCK DEFINITION
localparam Tclk = 10;
localparam Tclkd = (Tclk/2.0);
always @(*) #(Tclk/2.0) clk <= ~clk;

// REGS
reg clk, reset, start, end_pt, flag_perm, pt_ready, done_hard, last, rst, start_reseed, out_ready; 
reg [0:width*3-1] in;
reg [0:width*2-1] key;
reg [0:width*5*and_pini_nrnd-1] rnd;
reg [0:80-1] seed;
// WIRE
wire done, out_valid, flag_transition, busy;
wire [0:25*width-1] out_rnd;
wire [0:5*width-1] out_umsk;
wire [0:5*width*d-1] out;
wire [0:3*width*d-1] m_state;
wire [0:2*width*d-1] m_key;

prng_top #(.RND(25*width), .NINIT(4*288), .MAX_UNROLL(128)) trivium1 (
    reset,
    clk,            // Clock
    start_reseed,   // Acive high. Start the reseeding procedure.

    // Active high, reseed in progress,
    busy,
    // The seed, loaded as initial state.
    seed,
    // SVRS pseudo-random output stream.
    // Output will be always valid, except:
    // - when a reseeding is in progress
    // - when the PRNG has not been reseeded yet after a reset.
    out_ready,
    out_valid,
    out_rnd
    );

MSKcst #(.d(d), .count(3*width)) StateCst (in, m_state);
MSKcst #(.d(d), .count(2*width)) KeyCst (key, m_key);

reg [width*3-1:0] state_in0, state_in1;
reg [width*2-1:0] key_in0, key_in1;
wire [width*5*d-1:0] state_out;

top_prolead_ascon_hpc4 #(.width(width)) dut (
clk, 
start,
reset, 
state_in0,
state_in1, 
key_in0,
key_in1, 
rnd, 
state_out,
done
);

initial begin
	$dumpfile ("waveform.vcd");
    $dumpvars (0, DPA_tb_ascon);
    clk = 0;
    reset = 1;
    end_pt = 0;
    pt_ready = 1;
    flag_perm = 1;
    seed = 80'h0;
    state_in0 = {(width*3){2'b10}};
    state_in1 = {(width*3*d){1'b0}};
    key_in0 = {(width*2*d){1'b1}};
    key_in1 = {(width*2*d){1'b0}};

    out_ready = 1;
    rnd = {(64*5*and_pini_nrnd){1'b1}};
    last = 0;
    #40
    reset = 0;
    start = 1;
    in = {64'h80400c0600000000, {(128){1'b0}}};
    key = {128'b0};
    start_reseed = 1;
    #10
    start_reseed = 0;
    start = 0;
    //in = 320'hx;
    #40
    pt_ready = 1;
    #100
    start = 1;
    #400
    last = 1;
    end_pt = 1;
    #10
    end_pt = 0;
    $finish();
end


always@(done) begin
	if(done & last) begin
		#20
		$finish();
	end
end

// Debug
genvar j;
    generate
        for(j=0; j<5*width; j=j+1) begin
            assign out_umsk[j] = ^(out[d*j:d*(j+1)-1]);
        end
    endgenerate

endmodule