`timescale 1ns / 1ps
module DPA_Ascon_datapath #(parameter d = 2, width = 64) (clk, start, flag_dp_perm_a, state, rnd, state_rounded_csl);

parameter L = 16;
input clk, start, flag_dp_perm_a;
input [width*5*d-1:0] state;
input [width*25*d*(d-1)/2-1:0] rnd;

output [width*5*d-1:0] state_rounded_csl;

reg [width*5-1:0] outmux_f;
wire [width*5-1:0] outmux, u_state_csl, u_state_cs, u_state_c;
wire en;
wire[8-1:0] cst;
wire[8*d-1:0] m_cst;
wire [width*5*d-1:0] state_rounding, state_rounding_c, state_rounding_cs;

MSKmux #(.d(d), .count(width*5)) mux_state (start, state, state_rounded_csl, state_rounding);

dpa_rc_gen cstgeneration (clk, start, flag_dp_perm_a, cst);
// SHARING CREATION
MSKcst #(.d(d), .count(8)) msk_state (cst, m_cst);

// P_c
/*
assign state_rounding_c[2*width*d +: 8*d] = state_rounding[2*width*d +:8*d] ^ m_cst;
assign state_rounding_c[width*5*d-1:136*d] = state_rounding[5*width*d-1:136*d];
assign state_rounding_c[2*width*d-1 : 0] = state_rounding[2*width*d-1:0];
*/
assign state_rounding_c[0] = state_rounding ^ 1;
assign state_rounding_c[width*5*d-1:1] = state_rounding[width*5*d-1:1];

// P_s
    genvar i;
    generate 
    for(i=0; i<width; i = i + 1) begin: sboxs 
        MSK_Sbox #(.d(d)) sboxi (
            .clk(clk),
            .rnd(rnd[25*d*(d-1)/2*i +: 25*d*(d-1)/2]),
            .in({state_rounding_c[width*4*d + i*d +: d], state_rounding_c[width*3*d + i*d +: d], state_rounding_c[width*2*d + i*d +: d], state_rounding_c[width*d + i*d +: d], state_rounding_c[i*d +: d]}),
            .out({state_rounding_cs[width*4*d + i*d +: d], state_rounding_cs[width*3*d + i*d +: d], state_rounding_cs[width*2*d + i*d +: d], state_rounding_cs[width*d + i*d +: d], state_rounding_cs[i*d +: d]})
        );
    end
    endgenerate
    
// P_l
MSK_PermL #(.d(d), .width(width)) mskpermL (state_rounding_cs, state_rounded_csl);

`ifdef DEBUG
genvar j;
    generate
        for(j=0; j<5*width; j=j+1) begin
            assign u_state_c[j] = ^(state_rounding_c[d*(j+1)-1:d*j]);
            assign u_state_cs[j] = ^(state_rounding_cs[d*(j+1)-1:d*j]);
            assign u_state_csl[j] = ^(state_rounded_csl[d*(j+1)-1:d*j]);
        end
    endgenerate
`endif
endmodule