`timescale 1ns / 1ps
module MSKand_HPC4 #(parameter d=2) (ina, inb, rnd, clk, out);
localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam one_rand = d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [d-1:0] ina;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [d-1:0] inb;
(* syn_keep = "true", keep = "true", fv_type = "random", fv_count = 0, fv_rnd_lat_0 = 0, fv_rnd_count_0 = and_pini_nrnd *) input [and_pini_nrnd-1:0] rnd;
(* fv_type = "clock" *) input clk;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 1 *) output [d-1:0] out;

genvar i,j;

// unpack vector to matrix --> easier for randomness handling

wire [d-1:0] rnd_mat [d-1:0]; 
wire [d-1:0] rnd_mat1 [d-1:0]; 
wire [d-1:0] rnd_mat2 [d-1:0]; 
wire [d-1:0] rnd_mat3 [d-1:0]; 


for(i=0; i<d; i=i+1) begin: igen
    assign rnd_mat[i][i] = 0; 
    assign rnd_mat1[i][i] = 0;
    assign rnd_mat2[i][i] = 0;
    assign rnd_mat3[i][i] = 0;
    for(j=i+1; j<d; j=j+1) begin: jgen
        assign rnd_mat[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)];
        assign rnd_mat[i][j] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+d*(d-1)/2]; 

        assign rnd_mat1[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+2*d*(d-1)/2];
        assign rnd_mat1[i][j] = rnd_mat1[j][i];

        assign rnd_mat2[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+3*d*(d-1)/2]; 
        assign rnd_mat2[i][j] = rnd_mat2[j][i];

        assign rnd_mat3[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+4*d*(d-1)/2]; 
        assign rnd_mat3[i][j] = rnd_mat3[j][i];
    end
end

for(i=0; i<d; i=i+1) begin: ParProdI
    reg [d-2:0] u_F, v_F, w_F;
    reg aibi;
    reg ina_F;
    wire u_F_calc;
    always @(posedge clk) aibi <= ina[i] & inb[i];
    always @(posedge clk) ina_F <= ina[i];
    assign u_F_calc = ^u_F & ina_F;

    assign out[i] = aibi ^  u_F_calc ^ ^v_F ^ ^w_F;

    for(j=0; j<d; j=j+1) begin: ParProdJ
        if (i != j) begin: NotEq
            localparam j2 = j < i ?  j : j-1;

            wire u = inb[j] ^ rnd_mat[i][j] ^ rnd_mat1[i][j];
            wire v = ina[i] & rnd_mat[i][j] ^ rnd_mat2[i][j];
            wire w = ina[i] & rnd_mat1[i][j] ^ rnd_mat3[i][j];

            always @(posedge clk)
            begin
                u_F[j2] <= u;
                v_F[j2] <= v;
                w_F[j2] <= w;
            end
        end
    end
end

endmodule
