// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-P-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-P v2 (https://ohwr.org/cern_ohl_p_v2.txt).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-P v2 for applicable conditions.

`timescale 1ns / 1ps
module MSKand_HPC3plus #(parameter d=2, set_prng=0) (ina, inb, rnd, clk, out);
localparam and_pini_mul_nrnd = d*(d-1)+d;
localparam and_pini_nrnd = and_pini_mul_nrnd;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 1 *) input  [d-1:0] ina;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [d-1:0] inb;
(* syn_keep = "true", keep = "true", fv_type = "random", fv_count = 1, fv_rnd_lat_0 = 0, fv_rnd_count_0 = and_pini_nrnd *) input [and_pini_nrnd-1:0] rnd;
(* fv_type = "clock" *) input clk;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 2 *) output reg [d-1:0] out;

genvar i,j;

// unpack vector to matrix --> easier for randomness handling

wire [d-1:0] rnd_mat1 [d-1:0]; 
wire [d-1:0] rnd_mat2 [d-1:0]; 
wire [d-1:0] rnd_mat3; 

for(i=0; i<d; i=i+1) begin: igen
    assign rnd_mat1[i][i] = 0;
    assign rnd_mat2[i][i] = 0;
    for(j=i+1; j<d; j=j+1) begin: jgen
        assign rnd_mat1[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)];
        assign rnd_mat1[i][j] = rnd_mat1[j][i];
        assign rnd_mat2[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+d*(d-1)/2]; 
        assign rnd_mat2[i][j] = rnd_mat2[j][i];
        assign rnd_mat3[i] = rnd[d*(d-1)+i]; 
    end
end

wire [d-1:0] not_ina = ~ina;

for(i=0; i<d; i=i+1) begin: ParProdI
    reg [d-2:0] u, v;
    wire [d-2:0] c;
    reg aibi, ai_reg;
    wire aibi_comb = ina[i] & inb[i];
    always @(posedge clk) aibi <= aibi_comb;
    always @(posedge clk) ai_reg <= ina[i];
    always @(posedge clk ) out[i] = aibi ^ ^c ^ rnd_mat3[i];
        
    for(j=0; j<d; j=j+1) begin: ParProdJ
        if (i != j) begin: NotEq
            localparam j2 = j < i ?  j : j-1;
            wire u_j2_comb = inb[j] ^ rnd_mat1[i][j];
            wire v_j2_comb = not_ina[j] & rnd_mat1[i][j] ^ rnd_mat2[i][j];
            wire c_j2_comb = ai_reg & u[j2] ^ v[j2];
            assign c[j2] = c_j2_comb;
            always @(posedge clk)
            begin
                u[j2] <= u_j2_comb;
                v[j2] <= v_j2_comb;
            end
        end
    end
end

endmodule
