// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-P-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-P v2 (https://ohwr.org/cern_ohl_p_v2.txt).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-P v2 for applicable conditions.

`timescale 1ns / 1ps
module MSKand_HPC4_gated #(parameter d=2, set_prng=0) (ina, inb, rnd, sel_frwd, clk, out);
localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam one_rand = d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [0:d-1] ina;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 0 *) input  [0:d-1] inb;
(* syn_keep = "true", keep = "true", fv_type = "random", fv_count = 0, fv_rnd_lat_0 = 0, fv_rnd_count_0 = and_pini_nrnd *) input [and_pini_nrnd-1:0] rnd;
input sel_frwd;
(* fv_type = "clock" *) input clk;
(* syn_keep = "true", keep = "true", fv_type = "sharing", fv_latency = 1 *) output [0:d-1] out;
wire clk_g;
reg gating;
assign clk_g = gating | clk;

genvar i,j;

// unpack vector to matrix --> easier for randomness handling

wire [d-1:0] rnd_mat [d-1:0]; 
wire [d-1:0] rnd_mat1 [d-1:0]; 
wire [d-1:0] rnd_mat2 [d-1:0]; 
wire [d-1:0] rnd_mat3 [d-1:0]; 
wire [0:d-1] out_comp;
for(i=0; i<d; i=i+1) begin: igen
    assign rnd_mat[i][i] = 0; 
    assign rnd_mat1[i][i] = 0;
    assign rnd_mat2[i][i] = 0;
    assign rnd_mat3[i][i] = 0;
    for(j=i+1; j<d; j=j+1) begin: jgen
        assign rnd_mat[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)];
        assign rnd_mat[i][j] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+d*(d-1)/2]; 

        assign rnd_mat1[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+2*d*(d-1)/2];
        assign rnd_mat1[i][j] = rnd_mat1[j][i];

        assign rnd_mat2[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+3*d*(d-1)/2]; 
        assign rnd_mat2[i][j] = rnd_mat2[j][i];

        assign rnd_mat3[j][i] = rnd[((i*d)-i*(i+1)/2)+(j-1-i)+4*d*(d-1)/2]; 
        assign rnd_mat3[i][j] = rnd_mat3[j][i];

    end
end

for(i=0; i<d; i=i+1) begin: ParProdI
    reg [d-2:0] u_F, v_F, w_F;
    reg aibi;
    reg ina_F;
    wire u_F_calc;

    if (i == 0) begin 
        always @(posedge clk) aibi <= ina[i] & inb[i];
    end
    else begin
        always @(posedge clk_g) aibi <= ina[i] & inb[i];
    end

    always @(posedge clk_g) ina_F <= ina[i];
    always @(posedge clk) gating <= sel_frwd;
    assign u_F_calc = ^u_F & ina_F;

    assign out_comp[i] = aibi ^  u_F_calc ^ ^v_F ^ ^w_F;

    if (i == 0) begin 
        assign out[i] = (sel_frwd) ? aibi : out_comp[i];
    end
    else begin
        assign out[i] = (sel_frwd) ? 1'b0 : out_comp[i];
    end

    for(j=0; j<d; j=j+1) begin: ParProdJ
        if (i != j) begin: NotEq
            localparam j2 = j < i ?  j : j-1;

            wire u = inb[j] ^ rnd_mat[i][j] ^ rnd_mat1[i][j];
            wire v = ina[i] & rnd_mat[i][j] ^ rnd_mat2[i][j];
            wire w = ina[i] & rnd_mat1[i][j] ^ rnd_mat3[i][j];

            always @(posedge clk_g)
            begin
                u_F[j2] <= u;
                v_F[j2] <= v;
                w_F[j2] <= w;
            end
        end
    end
end

endmodule
