// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-P-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-P v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-P v2 (https://ohwr.org/cern_ohl_p_v2.txt).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-P v2 for applicable conditions.

// Masked register.
`timescale 1ns / 1ps
module MSKregGated #(parameter d=1, parameter count=1) (clk, sel, in, out);

(* fv_type = "clock" *)   input clk;
(* fv_type = "control" *)   input sel;
(* syn_keep="true", keep="true", fv_type = "sharing", fv_latency = 0, fv_count=count *) input  [0:count*d-1] in;
(* syn_keep="true", keep="true", fv_type = "sharing", fv_latency = 1, fv_count=count *) output [0:count*d-1] out;

(* syn_preserve="true", preserve="true" *) reg [0:count*d-1] state;
wire clk_g;
assign clk_g = clk | sel;
integer i;
always @(posedge clk) begin
    for(i=0; i<count; i=i+1) begin: i_ungated
        state[i*d] <= in[i*d];
    end
end

integer j;
always @(posedge clk_g) begin
    for(j=0; j<count; j=j+1) begin: i_gated
        state[j*d+1 +: d-1] <= in[j*d+1 +: d-1];
    end
end
assign out = state;
endmodule
