// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module keccak_FSM #(parameter d = 2, parameter set_prng=0, parameter depth = 64) (flag_start, flag_reset, clk, state_in, rnd, state_out, flag_done);

localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;

input flag_start, flag_reset, clk;
input [0:25*depth*d-1] state_in;
input [0:25*depth*and_pini_nrnd-1] rnd;
output [0:25*depth*d-1] state_out;
output reg flag_done;

// STATE DEFINITION

    `define IDLE 0
    `define COMPUTE 1

// Regs and wire

reg flag_dp_start, reset_round_cnt;
reg [0:1] state, next_state;
reg [0:4] round_cnt;

keccak_datapath #(.d(d), .depth(depth)) datapath (clk, flag_dp_start, state_in, rnd, state_out);

always @(*) begin
    reset_round_cnt = 0;
    flag_done = 0;
    flag_dp_start = 0;
    next_state = state;
    
    case(state)
        `IDLE : begin
            reset_round_cnt = 1;
            if(flag_start) begin
                flag_dp_start = 1; 
                reset_round_cnt = 1;
                next_state = `COMPUTE;
            end
        end
        `COMPUTE : begin
            if (round_cnt == 23) begin
                flag_done = 1;
                reset_round_cnt = 1;
                next_state = `IDLE;
            end
        end
    endcase
end

always @(posedge clk) begin
    if (flag_reset) begin
        state <=  `IDLE;
    end
    else begin
        state <= next_state;
    end
end 

always @(posedge clk) begin
    if(reset_round_cnt) begin
        round_cnt <= 0;
    end
    else begin
        round_cnt <= round_cnt + 1;
    end
end


endmodule