// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module keccak_tb ();
parameter d = 2;
localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
parameter depth = 64;
// Regs and wires 
reg start, reset, clk;
reg [0:25*depth*d-1] state_in;
reg [0: 25*depth*and_pini_nrnd-1] rnd;
wire done;
wire [0:25*depth*d-1] state_out;
// CLOCK DEFINITION
localparam Tclk = 10;
localparam Tclkd = (Tclk/2.0);
always @(*) #(Tclk/2.0) clk <= ~clk;

keccak_FSM #(.d(d), .depth(depth)) dut (start, reset, clk, state_in, rnd, state_out, done);

initial begin
	$dumpfile (`VCD_FILE);
    $dumpvars (0, keccak_tb);
    clk = 0;
    start = 0;
    reset = 1;
    state_in = {4'b1100, {(399){4'b1110}}};
    rnd = 1;
    #10
    start = 1;
    reset = 0;
    #50
    $finish();
end

endmodule