// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module ketje_FSM #(parameter d = 2, parameter set_prng=0, parameter depth = 16) (flag_start, flag_reset, flag_last_ad, flag_last_pt, clk, key, nonce, plaintext, rnd, ciphertext, tag, flag_done);

localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;

input flag_start, flag_reset, flag_last_ad, flag_last_pt, clk;
input [0:128*d-1] key, nonce;
input [0:32*d-1] plaintext;
input [0:25*depth*and_pini_nrnd-1] rnd;
output [0:32*d-1] ciphertext;
output [0:128*d-1] tag;
output reg flag_done;

// STATE DEFINITION

    `define IDLE 0
    `define INIT 1
    `define AD 2
    `define PT 3
    `define T 4

// Regs and wire

reg [0:25*d*depth-1] state_in;

reg flag_dp_start, reset_round_cnt, flag_pt;
reg [0:2] state, next_state;
reg [0:4] round_cnt;

ketje_datapath #(.d(d), .depth(depth)) datapath (clk, flag_dp_start, state_in, plaintext, flag_pt, rnd, ciphertext, tag);

always @(*) begin
    reset_round_cnt = 0;
    flag_done = 0;
    next_state = state;
    flag_dp_start = 0;
    flag_pt = 0;
    state_in = 0;
    case(state)
        `IDLE : begin
            reset_round_cnt = 1;
            if(flag_start) begin
                flag_dp_start = 1; 
                state_in = {key, nonce};
                reset_round_cnt = 1;
                next_state = `INIT;
            end
        end
        `INIT : begin
            if (round_cnt == 11) begin
                reset_round_cnt = 1;
                next_state = `AD;
            end
        end
        `AD : begin
            if (round_cnt == 0 & flag_last_ad) begin
                next_state = `PT;
                reset_round_cnt = 1;
            end
            else if (round_cnt == 0) begin
                reset_round_cnt = 1;
            end
        end
        `PT : begin
            if (round_cnt == 0 & flag_last_pt) begin
                reset_round_cnt = 1;
                flag_pt = 1;
                next_state = `T;
            end
            else if (round_cnt == 0) begin
                reset_round_cnt = 1;
                flag_pt = 1;;
            end
        end
        `T : begin
            if (round_cnt == 6) begin
                next_state = `IDLE;
                flag_done = 1;
                reset_round_cnt = 1;
            end
        end
    endcase
end

always @(posedge clk) begin
    if (flag_reset) begin
        state <=  `IDLE;
    end
    else begin
        state <= next_state;
    end
end 

always @(posedge clk) begin
    if(reset_round_cnt) begin
        round_cnt <= 0;
    end
    else begin
        round_cnt <= round_cnt + 1;
    end
end


endmodule