// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

`timescale 1ns / 1ps
module ketje_datapath  #(parameter d = 2, parameter depth = 16) (clk, start, state_in, plaintext, flag_pt, rnd, ciphertext, tag);

localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
localparam matrix = 25;

input clk, start;
input [0:25*depth*d-1] state_in;
input [0:32*d-1] plaintext;
input flag_pt;
input [0:25*depth*and_pini_nrnd-1] rnd;
output [0:32*d-1] ciphertext;
output [0:128*d-1] tag;

wire [0:25*depth*d-1] debug, debug2;
wire[0:depth*d-1] lane, lane_out, rconst;
wire [0:4] row, row_out;

wire [0:depth*d-1] chi_in [0:4] [0:4]; 
wire [0:depth*d-1] chi_out [0:4] [0:4]; 
wire [0:depth*d-1] inved1 [0:4] [0:4]; 
wire [0:depth*d-1] anded1 [0:4] [0:4]; 
wire [0:depth*d-1] chi_in_F [0:4] [0:4]; 

wire [0:depth*d-1] theta_in [0:4] [0:4]; 
wire [0:depth*d-1] theta_out [0:4] [0:4]; 

wire [0:depth*d-1] pi_in [0:4] [0:4]; 
wire [0:depth*d-1] pi_out [0:4] [0:4]; 

wire sum_sheet [0:4] [0:depth*d-1];
wire sum_0 [0:4] [0:depth*d-1];
wire [0:depth*d-1] rho_in [0:4] [0:4]; 
wire [0:depth*d-1] rho_out [0:4] [0:4]; 

wire [0:depth*d-1] iota_in [0:4] [0:4]; 
wire [0:depth*d-1] iota_out [0:4] [0:4]; 

wire [0:depth*d-1] rounded [0:4] [0:4]; 
wire [0:depth*d-1] rounded_abs [0:4] [0:4]; 

wire [0:depth*d-1] mux_f_state_in [0:4] [0:4]; 
wire [0:depth*d-1] format_state_in [0:4] [0:4]; 

wire [0:32*d-1] plaintext_mux;

genvar x, y, z;

MSKmux #(.d(d), .count(32)) mux_pt (flag_pt, plaintext, {(32*d){1'b0}}, plaintext_mux);
MSKxor #(.d(d), .count(depth)) xor_pt0 (plaintext_mux[0:depth*d-1], rounded[0][0], rounded_abs[0][0]);
MSKxor #(.d(d), .count(depth)) xor_pt1 (plaintext_mux[depth*d:2*depth*d-1], rounded[0][1], rounded_abs[0][1]);

assign ciphertext = {rounded_abs[4][4], rounded_abs[4][3]};



assign rounded_abs[0][2] = rounded[0][2];
assign rounded_abs[0][3] = rounded[0][3];
assign rounded_abs[0][4] = rounded[0][4];
generate
    for (x = 1; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign rounded_abs[x][y] = rounded[x][y];
        end
    end
endgenerate

generate
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            MSKmux #(.d(d), .count(depth)) mux_state (start, format_state_in[x][y], rounded[x][y], mux_f_state_in[x][y]);
        end
    end
endgenerate
// Pack the vector as a depthx5x5 matrix
generate
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            for (z = 0; z < depth*d; z = z + 1) begin
                assign format_state_in[x][y][z] = state_in[z+d*depth*y+d*5*depth*x];
            end
        end
    end
endgenerate


generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign theta_in[x][y] = mux_f_state_in[x][y];            
        end
    end            
endgenerate

generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign rho_in[x][y] = theta_out[x][y];            
        end
    end            
endgenerate

generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign pi_in[x][y] = rho_out[x][y];            
        end
    end            
endgenerate

generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign chi_in[x][y] = pi_out[x][y];            
        end
    end            
endgenerate

generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign iota_in[x][y] = chi_out[x][y];            
        end
    end            
endgenerate

// DEBUG
generate
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            for (z = 0; z < depth*d; z = z + 1) begin
                assign debug[x*5*depth*d+y*depth*d+z] = format_state_in[x][y][z];
                assign debug2[x*5*depth*d+y*depth*d+z] = theta_out[x][y][z];
            end
        end
    end
endgenerate

generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            assign rounded[x][y] = iota_out[x][y];            
        end
    end            
endgenerate

// Perform the chi operation

generate
    for (x = 0; x < 3; x = x + 1) begin
        for (y = 0; y < 5; y = y + 1) begin
            MSKreg #(.d(d), .count(depth)) reg1 (clk, chi_in[x][y], chi_in_F[x][y]);
            for (z = 0; z < depth; z = z + 1) begin
                //assign chi_out[x][y][z] = chi_in[x][y][z] ^ (~chi_in[x+1][y][z] & chi_in[x+2][y][z]);

                MSKinv #(.d(d), .count(1)) inv_1 (chi_in[x+1][y][z*d:(z+1)*d-1], inved1[x][y][z*d:(z+1)*d-1]);
                MSKand_HPC4 #(.d(d)) and1 (inved1[x][y][z*d:(z+1)*d-1], chi_in[x+2][y][z*d:(z+1)*d-1], rnd[z*and_pini_nrnd +: and_pini_nrnd], clk, anded1[x][y][z*d:(z+1)*d-1]);
                MSKxor #(.d(d), .count(1)) xor1 (anded1[x][y][z*d:(z+1)*d-1], chi_in_F[x][y][z*d:(z+1)*d-1], chi_out[x][y][z*d:(z+1)*d-1]);
            end
        end
    end
endgenerate

generate
    for (y = 0; y < 5; y = y + 1) begin
        MSKreg #(.d(d), .count(depth)) reg1 (clk, chi_in[3][y], chi_in_F[3][y]);
        for (z = 0; z < depth; z = z + 1) begin
            //assign chi_out[3][y][z] = chi_in[3][y][z] ^ (~chi_in[4][y][z] & chi_in[0][y][z]);

            MSKinv #(.d(d), .count(1)) inv_2 (chi_in[4][y][z*d:(z+1)*d-1], inved1[3][y][z*d:(z+1)*d-1]);
            MSKand_HPC4 #(.d(d)) and2 (inved1[3][y][z*d:(z+1)*d-1], chi_in[0][y][z*d:(z+1)*d-1], rnd[z*and_pini_nrnd +: and_pini_nrnd], clk, anded1[3][y][z*d:(z+1)*d-1]);
            MSKxor #(.d(d), .count(1)) xor2 (anded1[3][y][z*d:(z+1)*d-1], chi_in_F[3][y][z*d:(z+1)*d-1], chi_out[3][y][z*d:(z+1)*d-1]);

        end
    end
endgenerate

generate
    for (y = 0; y < 5; y = y + 1) begin
            MSKreg #(.d(d), .count(depth)) reg1 (clk, chi_in[0][y], chi_in_F[4][y]);

            for (z = 0; z < depth; z = z + 1) begin
            //assign chi_out[4][y][z] = chi_in[4][y][z] ^ (~chi_in[0][y][z] & chi_in[1][y][z]);

            MSKinv #(.d(d), .count(1)) inv_3 (chi_in[0][y][z*d:(z+1)*d-1], inved1[4][y][z*d:(z+1)*d-1]);
            MSKand_HPC4 #(.d(d)) and3 (inved1[4][y][z*d:(z+1)*d-1], chi_in[0][y][z*d:(z+1)*d-1], rnd[z*and_pini_nrnd +: and_pini_nrnd], clk, anded1[4][y][z*d:(z+1)*d-1]);
            MSKxor #(.d(d), .count(1)) xor3 (anded1[4][y][z*d:(z+1)*d-1], chi_in_F[4][y][z*d:(z+1)*d-1], chi_out[4][y][z*d:(z+1)*d-1]);
        end
    end
endgenerate

// Theta

//compute the sum of the columns

generate
    for(x = 0; x < 5; x = x+1) begin
        for(z = 0; z < depth*d; z = z + 1) begin
            assign sum_sheet[x][z] = theta_in[x][0][z] ^ theta_in[x][1][z] ^ theta_in[x][2][z] ^ theta_in[x][3][z] ^ theta_in[x][4][z];
            //MSKxor #(.d(d), .count(1)) xor_D0 (theta_in[x][0][z*d:(z+1)*d-1], theta_in[x][1][z*d:(z+1)*d-1], sum_0[x][z*d:(z+1)*d-1]);
        end
    end
endgenerate

// compute for x = 1,2,3 and z = 1,...,63 and all y
generate
    for(x = 1; x < 4; x = x +1)
        for(y = 0; y < 5; y = y + 1) begin
            assign theta_out[x][y][0] = theta_in[x][y][0] ^ sum_sheet[x-1][0] ^ sum_sheet[x+1][depth*d-1];

            for(z = 1; z < depth*d; z = z+1)
                assign theta_out[x][y][z] = theta_in[x][y][z] ^ sum_sheet[x-1][z] ^ sum_sheet[x+1][z-1];
        end
endgenerate

// compute for x = 0 and z = 1,...,63 and all y
generate
    for(y = 0; y < 5; y = y+1) begin
        assign theta_out[0][y][0] = theta_in[0][y][0] ^ sum_sheet[4][0] ^ sum_sheet[1][depth*d-1];

        for(z = 1; z < depth*d; z=z+1)
            assign theta_out[0][y][z] = theta_in[0][y][z] ^ sum_sheet[4][z] ^ sum_sheet[1][z-1];
    end
endgenerate
// compute for x = 4 and z = 1,...,63 and all y
generate
    for(y = 0; y < 5; y = y+1) begin
        assign theta_out[4][y][0] = theta_in[4][y][0] ^ sum_sheet[3][0] ^ sum_sheet[0][depth*d-1];

        for(z = 1; z < depth*d; z=z+1)
            assign theta_out[4][y][z] = theta_in[4][y][z] ^ sum_sheet[3][z] ^ sum_sheet[0][z-1];
    end
endgenerate


// Pi
// hardcode mapping

assign pi_out[0][0] = pi_in[0][0];
assign pi_out[1][0] = pi_in[0][3];
assign pi_out[2][0] = pi_in[0][1];
assign pi_out[3][0] = pi_in[0][4];
assign pi_out[4][0] = pi_in[0][2];

assign pi_out[0][1] = pi_in[1][1];
assign pi_out[1][1] = pi_in[1][4];
assign pi_out[2][1] = pi_in[1][2];
assign pi_out[3][1] = pi_in[1][0];
assign pi_out[4][1] = pi_in[1][3];

assign pi_out[0][2] = pi_in[2][2];
assign pi_out[1][2] = pi_in[2][0];
assign pi_out[2][2] = pi_in[2][3];
assign pi_out[3][2] = pi_in[2][1];
assign pi_out[4][2] = pi_in[2][4];

assign pi_out[0][3] = pi_in[3][3];
assign pi_out[1][3] = pi_in[3][1];
assign pi_out[2][3] = pi_in[3][4];
assign pi_out[3][3] = pi_in[3][2];
assign pi_out[4][3] = pi_in[3][0];

assign pi_out[0][4] = pi_in[4][4];
assign pi_out[1][4] = pi_in[4][2];
assign pi_out[2][4] = pi_in[4][0];
assign pi_out[3][4] = pi_in[4][3];
assign pi_out[4][4] = pi_in[4][1];

// Rho
// hardcode mapping

assign lane = rho_in[1][0];
assign lane_out = rho_out[1][0];

/*
 KECCAK 1600
assign rho_out[0][0] = rho_in[0][0];
assign rho_out[1][0] = {rho_in[1][0][63*d:depth*d-1], rho_in[1][0][0:63*d-1]};
assign rho_out[2][0] = {rho_in[2][0][2*d:depth*d-1], rho_in[2][0][0:2*d-1]};
assign rho_out[3][0] = {rho_in[3][0][36*d:depth*d-1], rho_in[3][0][0:36*d-1]};
assign rho_out[4][0] = {rho_in[4][0][35*d:depth*d-1], rho_in[4][0][0:35*d-1]};

assign rho_out[0][1] = {rho_in[0][1][27*d:depth*d-1], rho_in[1][1][0:28*d-1]};
assign rho_out[1][1] = {rho_in[1][1][20*d:depth*d-1], rho_in[1][1][0:20*d-1]};
assign rho_out[2][1] = {rho_in[2][1][58*d:depth*d-1], rho_in[2][1][0:58*d-1]};
assign rho_out[3][1] = {rho_in[3][1][9*d:depth*d-1], rho_in[3][1][0:9*d-1]};
assign rho_out[4][1] = {rho_in[4][1][44*d:depth*d-1], rho_in[4][1][0:44*d-1]};

assign rho_out[0][2] = {rho_in[0][2][61*d:depth*d-1], rho_in[0][2][0:61*d-1]};
assign rho_out[1][2] = {rho_in[1][2][53*d:depth*d-1], rho_in[1][2][0:54*d-1]};
assign rho_out[2][2] = {rho_in[2][2][21*d:depth*d-1], rho_in[2][2][0:21*d-1]};
assign rho_out[3][2] = {rho_in[3][2][38*d:depth*d-1], rho_in[3][2][0:39*d-1]};
assign rho_out[4][2] = {rho_in[4][2][25*d:depth*d-1], rho_in[4][2][0:25*d-1]};

assign rho_out[0][3] = {rho_in[0][3][23*d:depth*d-1], rho_in[0][3][0:23*d-1]};
assign rho_out[1][3] = {rho_in[1][3][19*d:depth*d-1], rho_in[1][3][0:19*d-1]};
assign rho_out[2][3] = {rho_in[2][3][21*d:depth*d-1], rho_in[2][3][0:21*d-1]};
assign rho_out[3][3] = {rho_in[3][3][43*d:depth*d-1], rho_in[3][3][0:43*d-1]};
assign rho_out[4][3] = {rho_in[4][3][56*d:depth*d-1], rho_in[4][3][0:56*d-1]};

assign rho_out[0][4] = {rho_in[0][4][46*d:depth*d-1], rho_in[0][4][0:46*d-1]};
assign rho_out[1][4] = {rho_in[1][4][62*d:depth*d-1], rho_in[1][4][0:62*d-1]};
assign rho_out[2][4] = {rho_in[2][4][3*d:depth*d-1], rho_in[2][4][0:3*d-1]};
assign rho_out[3][4] = {rho_in[3][4][8*d:depth*d-1], rho_in[3][4][0:8*d-1]};
assign rho_out[4][4] = {rho_in[4][4][50*d:depth*d-1], rho_in[4][4][0:50*d-1]};
*/

// KETJE 400
assign rho_out[0][0] = rho_in[0][0];
assign rho_out[1][0] = {rho_in[1][0][1*d:depth*d-1], rho_in[1][0][0:1*d-1]};
assign rho_out[2][0] = {rho_in[2][0][14*d:depth*d-1], rho_in[2][0][0:14*d-1]};
assign rho_out[3][0] = {rho_in[3][0][12*d:depth*d-1], rho_in[3][0][0:12*d-1]};
assign rho_out[4][0] = {rho_in[4][0][11*d:depth*d-1], rho_in[4][0][0:11*d-1]};

assign rho_out[0][1] = {rho_in[0][1][4*d:depth*d-1], rho_in[1][1][0:4*d-1]};
assign rho_out[1][1] = {rho_in[1][1][12*d:depth*d-1], rho_in[1][1][0:12*d-1]};
assign rho_out[2][1] = {rho_in[2][1][6*d:depth*d-1], rho_in[2][1][0:6*d-1]};
assign rho_out[3][1] = {rho_in[3][1][7*d:depth*d-1], rho_in[3][1][0:7*d-1]};
assign rho_out[4][1] = {rho_in[4][1][4*d:depth*d-1], rho_in[4][1][0:4*d-1]};

assign rho_out[0][2] = {rho_in[0][2][3*d:depth*d-1], rho_in[0][2][0:3*d-1]};
assign rho_out[1][2] = {rho_in[1][2][1*d:depth*d-1], rho_in[1][2][0:1*d-1]};
assign rho_out[2][2] = {rho_in[2][2][11*d:depth*d-1], rho_in[2][2][0:11*d-1]};
assign rho_out[3][2] = {rho_in[3][2][9*d:depth*d-1], rho_in[3][2][0:9*d-1]};
assign rho_out[4][2] = {rho_in[4][2][7*d:depth*d-1], rho_in[4][2][0:7*d-1]};

assign rho_out[0][3] = {rho_in[0][3][9*d:depth*d-1], rho_in[0][3][0:9*d-1]};
assign rho_out[1][3] = {rho_in[1][3][13*d:depth*d-1], rho_in[1][3][0:13*d-1]};
assign rho_out[2][3] = {rho_in[2][3][15*d:depth*d-1], rho_in[2][3][0:15*d-1]};
assign rho_out[3][3] = {rho_in[3][3][5*d:depth*d-1], rho_in[3][3][0:5*d-1]};
assign rho_out[4][3] = {rho_in[4][3][8*d:depth*d-1], rho_in[4][3][0:8*d-1]};

assign rho_out[0][4] = {rho_in[0][4][8*d:depth*d-1], rho_in[0][4][0:8*d-1]};
assign rho_out[1][4] = {rho_in[1][4][14*d:depth*d-1], rho_in[1][4][0:14*d-1]};
assign rho_out[2][4] = {rho_in[2][4][2*d:depth*d-1], rho_in[2][4][0:2*d-1]};
assign rho_out[3][4] = {rho_in[3][4][2*d:depth*d-1], rho_in[3][4][0:2*d-1]};
assign rho_out[4][4] = {rho_in[4][4][13*d:depth*d-1], rho_in[4][4][0:13*d-1]};

generate 
    for (x = 0; x < 5; x = x + 1) begin
        for (y = 1; y < 5; y = y + 1) begin
            assign iota_out[x][y] = iota_in[x][y];            
        end
    end            
endgenerate

generate 
    for (x = 1; x < 5; x = x + 1) begin
        assign iota_out[x][0] = iota_in[x][0];            
    end            
endgenerate

rc_gen #(.d(d), .depth(depth)) round_constant_gen (clk, start, rconst);

assign iota_out[0][0] = iota_in[0][0] ^ rconst;

endmodule