// SPDX-FileCopyrightText: SIMPLE-Crypto Contributors <info@simple-crypto.dev>
// SPDX-License-Identifier: CERN-OHL-S-2.0
// Copyright SIMPLE-Crypto Contributors.
// This source describes Open Hardware and is licensed under the CERN-OHL-S v2.
// You may redistribute and modify this source and make products using it under
// the terms of the CERN-OHL-S v2 (https://ohwr.org/cern_ohl_s_v2.txt ).
// This source is distributed WITHOUT ANY EXPRESS OR IMPLIED WARRANTY, INCLUDING
// OF MERCHANTABILITY, SATISFACTORY QUALITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Please see the CERN-OHL-S v2 for applicable conditions.
// Source location: https://github.com/simple-crypto/SMAesH
// As per CERN-OHL-S v2 section 4, should You produce hardware based
// on this source, You must where practicable maintain the Source Location
// visible on the external case of any product you make using this source.

module keccak_tb ();
parameter d = 2;
localparam and_pini_mul_nrnd = 5*d*(d-1)/2;
localparam and_pini_nrnd = and_pini_mul_nrnd;
parameter depth = 16;
// Regs and wires 
reg start, reset, clk, flag_last_ad, flag_last_pt;
reg [0:128*d-1] key, nonce;
reg [0:32*d-1] plaintext;
reg [0: 25*depth*and_pini_nrnd-1] rnd;
wire done;
wire [0:32*d-1] ciphertext;
wire [0:128*d-1] tag;
// CLOCK DEFINITION
localparam Tclk = 10;
localparam Tclkd = (Tclk/2.0);
always @(*) #(Tclk/2.0) clk <= ~clk;

ketje_FSM_d2 #(.d(d), .depth(16)) dut (start, reset, flag_last_ad, flag_last_pt, clk, key, nonce, plaintext, rnd, ciphertext, tag, done);

initial begin
	$dumpfile (`VCD_FILE);
    $dumpvars (0, keccak_tb);
    clk = 0;
    start = 0;
    reset = 1;
    key = 0;
    nonce = 0;
    plaintext = 0;
    flag_last_ad = 0;
    flag_last_pt = 0;
    rnd = 1;
    #10
    start = 1;
    reset = 0;
    #50
    flag_last_ad = 1;
    flag_last_pt = 1;
    #1000
    $finish();
end

endmodule