set PARAM_D $::env(PARAM_D)
set PARAM_PRNG 0

set_db message:PHYS/PHYS-15 .max_print 3
set_db message:PHYS/PHYS-129 .max_print 3
set_db message:LBR/LBR-9 .max_print 3
set_db message:LBR/LBR-76 .max_print 3

set DESIGN_NAME MSKand_HPC4_gated
set SDC_FILE setup1.sdc

set_db gui_sv_threshold 50000

set DEST_DIR $::env(DEST_DIR)

##########################################################################

set_db library "../../../freepdk-45nm/stdcells.lib"
set_db lef_library "../../../freepdk-45nm/rtk-tech.lef"
set_db cap_table_file "../../../freepdk-45nm/rtk-typical.captable"

##########################################################################
set_db delete_unloaded_insts true 
# PROPER ASSIGN STATEMENT ON OUTPUTS
set_db remove_assigns true

# RISE ERRORS IF ANY LATCH IS DETECTED
set_db hdl_error_on_latch true

#READ ALL DESIGN FILES
# Library
set_db init_hdl_search_path {"./hdl_cleaned/lib"}
read_hdl -sv {
MSKinv.v \
MSKcst.v \
MSKmux.v \
MSKregGated.v \
MSKregEn.v \
MSKreg.v \
regEn.v \
MSKand_HPC4_gated.v \
MSKxor.v \
prng_top.v \
trivium_prng.v
}
# prng 
set_db init_hdl_search_path {"./hdl_cleaned/prng"}
read_hdl -sv {
prng_top.v \
trivium_prng.v
}
# Ascon dpa core
set_db init_hdl_search_path {"./hdl_cleaned/dpa"}
read_hdl -sv {
DPA_Ascon_datapath.v \
DPA_Ascon_FSM.v \
DPA_rc_gen.v \
MSK_PermL.v \
MSK_Sbox.v
}
# Ascon spa core
set_db init_hdl_search_path {"./hdl_cleaned/spa"}
read_hdl -sv {
SPA_rc_gen.v \
PermL.v \
Sbox.v \
SPA_Ascon_datapath.v \
SPA_Ascon_FSM.v \
}
# Ascon mode core
set_db init_hdl_search_path {"./hdl_cleaned/mode"}
read_hdl -sv {
MSK_tgf_in.v \
MSK_tgf_out.v \
MSK_kdf.v \
MSK_stall_abs.v \
UMSK_stall_abs.v \
levelled_datapath.v \
levelled_FSM.v
}
##########################################################################

#RTL POWER ANALYSIS
set_db hdl_track_filename_row_col true 
set_db lp_power_unit mW 

#CLOCK GATING
set_db lp_insert_clock_gating true

#ELABORATE THE DESIGN and check
set parameters {}
lappend parameters [list d $PARAM_D]
lappend parameters [list set_prng $PARAM_PRNG]

set design [elaborate $DESIGN_NAME -parameters $parameters]
check_design -unresolved

set_db [get_db modules module:levelled_FSM*/levelled_datapath*] .ungroup_ok false
set_db [get_db modules module:levelled_FSM*/prng_top*] .ungroup_ok false

set_db [get_db modules MSK*] .dont_touch map_size_ok
set_db [get_db modules DPA*] .ungroup_ok false
set_db [get_db modules SPA*] .ungroup_ok false
set_db [get_db modules MSK*] .retime false

set_db [get_db $design .hinsts MSK*] .ungroup_ok false
set_db [get_db modules MSK*] .dont_touch map_size_ok
set_db [get_db modules MSK*] .retime false

#SETTING THE POWER OPTIMIZATION
set_db lp_power_analysis_effort high 
set_db lp_clock_gating_infer_enable true
set_db lp_insert_discrete_clock_gating_logic true
set_db lp_clock_gating_prefix CG_

#READ SDC
read_sdc setup1.sdc

#SYNTHESIZE THE DESIGN
set_db syn_global_effort high
syn_generic

write_snapshot -outdir $DEST_DIR/snapshot -tag generic
report_summary -directory $DEST_DIR/snapshot
# checkpoint
write_db -to_file $DEST_DIR/synth/design.db

syn_map
write_snapshot -outdir $DEST_DIR/snapshot -tag mapped
report_summary -directory $DEST_DIR/snapshot
syn_opt
write_snapshot -outdir $DEST_DIR/snapshot -tag optimized
report_summary -directory $DEST_DIR/snapshot

#REPORT TIMING
report timing > $DEST_DIR/report/report_timing.rep
report clocks > $DEST_DIR/report/report_clocks.rep
report_clock_gating -detail > $DEST_DIR/report/report_clock_gating.rep

#REPORT AREA
report_area > $DEST_DIR/report/report_area.rep

#REPORT GATES
report gates > $DEST_DIR/report/report_cell.rep

#REPORT POWER
report power -verbose > $DEST_DIR/report/report_power.rep

#WRITE BACK THE SYNTHESIZED NETLIST
write -mapped > $DEST_DIR/synth/design.v

#WRITE BACK THE DB and SDC FOR RECALLING
write_db -to_file $DEST_DIR/synth/design.db
write_design -gzip -basename $DEST_DIR/design/mapped
write_sdc > $DEST_DIR/synth/design.sdc

